# Copyright (c) 2020-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, Dict, Tuple
import math
from ezdxf.math import (
    Vec3,
    NULLVEC,
    X_AXIS,
    Z_AXIS,
    OCS,
    Matrix44,
    arc_angle_span_rad,
    linspace,
    distance_point_line_3d,
    enclosing_angles,
)

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex, BaseLayout, Ellipse

__all__ = [
    "ConstructionEllipse",
    "angle_to_param",
    "param_to_angle",
    "rytz_axis_construction",
]
QUARTER_PARAMS = [0, math.pi * 0.5, math.pi, math.pi * 1.5]
HALF_PI = math.pi / 2.0


class ConstructionEllipse:
    """This is a helper class to create parameters for 3D ellipses.

    Args:
        center: 3D center point
        major_axis: major axis as 3D vector
        extrusion: normal vector of ellipse plane
        ratio: ratio of minor axis to major axis
        start_param: start param in radians
        end_param: end param in radians
        ccw: is counter clockwise flag - swaps start- and end param if ``False``

    """

    def __init__(
        self,
        center: "Vertex" = NULLVEC,
        major_axis: "Vertex" = X_AXIS,
        extrusion: "Vertex" = Z_AXIS,
        ratio: float = 1,
        start_param: float = 0,
        end_param: float = math.tau,
        ccw: bool = True,
    ):
        self.center = Vec3(center)
        self.major_axis = Vec3(major_axis)
        if self.major_axis.isclose(NULLVEC):
            raise ValueError(f"Invalid major axis (null vector).")
        self.extrusion = Vec3(extrusion)
        if self.major_axis.isclose(NULLVEC):
            raise ValueError(f"Invalid extrusion vector (null vector).")
        self.ratio = float(ratio)
        self.start_param = float(start_param)
        self.end_param = float(end_param)
        if not ccw:
            self.start_param, self.end_param = self.end_param, self.start_param
        self.minor_axis = minor_axis(
            self.major_axis, self.extrusion, self.ratio
        )

    @classmethod
    def from_arc(
        cls,
        center: "Vertex" = NULLVEC,
        radius: float = 1,
        extrusion: "Vertex" = Z_AXIS,
        start_angle: float = 0,
        end_angle: float = 360,
        ccw: bool = True,
    ) -> "ConstructionEllipse":
        """Returns :class:`ConstructionEllipse` from arc or circle.

        Arc and Circle parameters defined in OCS.

        Args:
             center: center in OCS
             radius: arc or circle radius
             extrusion: OCS extrusion vector
             start_angle: start angle in degrees
             end_angle: end angle in degrees
             ccw: arc curve goes counter clockwise from start to end if ``True``
        """
        radius = abs(radius)
        if NULLVEC.isclose(extrusion):
            raise ValueError(f"Invalid extrusion: {str(extrusion)}")
        ratio = 1.0
        ocs = OCS(extrusion)
        center = ocs.to_wcs(center)
        # Major axis along the OCS x-axis.
        major_axis = ocs.to_wcs(Vec3(radius, 0, 0))
        # No further adjustment of start- and end angle required.
        start_param = math.radians(start_angle)
        end_param = math.radians(end_angle)
        return cls(
            center,
            major_axis,
            extrusion,
            ratio,
            start_param,
            end_param,
            bool(ccw),
        )

    def __copy__(self):
        return self.__class__(
            self.center,
            self.major_axis,
            self.extrusion,
            self.ratio,
            self.start_param,
            self.end_param,
        )

    @property
    def start_point(self) -> Vec3:
        """Returns start point of ellipse as Vec3."""
        return vertex(
            self.start_param,
            self.major_axis,
            self.minor_axis,
            self.center,
            self.ratio,
        )

    @property
    def end_point(self) -> Vec3:
        """Returns end point of ellipse as Vec3."""
        return vertex(
            self.end_param,
            self.major_axis,
            self.minor_axis,
            self.center,
            self.ratio,
        )

    def dxfattribs(self) -> Dict:
        """Returns required DXF attributes to build an ELLIPSE entity.

        Entity ELLIPSE has always a ratio in range from 1e-6 to 1.

        """
        if self.ratio > 1:
            e = self.__copy__()
            e.swap_axis()
        else:
            e = self
        return {
            "center": e.center,
            "major_axis": e.major_axis,
            "extrusion": e.extrusion,
            "ratio": max(e.ratio, 1e-6),
            "start_param": e.start_param,
            "end_param": e.end_param,
        }

    def main_axis_points(self) -> Iterable[Vec3]:
        """Yields main axis points of ellipse in the range from start- to end
        param.
        """
        start = self.start_param
        end = self.end_param
        for param in QUARTER_PARAMS:
            if enclosing_angles(param, start, end):
                yield vertex(
                    param,
                    self.major_axis,
                    self.minor_axis,
                    self.center,
                    self.ratio,
                )

    def transform(self, m: Matrix44) -> None:
        """Transform ellipse in place by transformation matrix `m`."""
        new_center = m.transform(self.center)
        # 2021-01-28 removed % math.tau
        old_start_param = start_param = self.start_param
        old_end_param = end_param = self.end_param
        old_minor_axis = minor_axis(self.major_axis, self.extrusion, self.ratio)
        new_major_axis, new_minor_axis = m.transform_directions(
            (self.major_axis, old_minor_axis)
        )
        # Original ellipse parameters stay untouched until end of transformation
        dot_product = new_major_axis.normalize().dot(new_minor_axis.normalize())
        if abs(dot_product) > 1e-6:
            new_major_axis, new_minor_axis, new_ratio = rytz_axis_construction(
                new_major_axis, new_minor_axis
            )
            new_extrusion = new_major_axis.cross(new_minor_axis).normalize()
            adjust_params = True
        else:
            # New axis are nearly orthogonal:
            new_ratio = new_minor_axis.magnitude / new_major_axis.magnitude
            # New normal vector:
            new_extrusion = new_major_axis.cross(new_minor_axis).normalize()
            # Calculate exact minor axis:
            new_minor_axis = minor_axis(
                new_major_axis, new_extrusion, new_ratio
            )
            adjust_params = False

        if adjust_params and not math.isclose(
            start_param, end_param, abs_tol=1e-9
        ):
            # open ellipse, adjusting start- and end parameter
            x_axis = new_major_axis.normalize()
            y_axis = new_minor_axis.normalize()
            # TODO: use ellipse_param_span()?
            #  2021-01-28 this is possibly the source of errors!
            old_param_span = (end_param - start_param) % math.tau

            def param(vec: "Vec3") -> float:
                dy = y_axis.dot(vec) / new_ratio  # adjust to circle
                dx = x_axis.dot(vec)
                return math.atan2(dy, dx) % math.tau

            # transformed start- and end point of old ellipse
            start_point = m.transform(
                vertex(
                    start_param,
                    self.major_axis,
                    old_minor_axis,
                    self.center,
                    self.ratio,
                )
            )
            end_point = m.transform(
                vertex(
                    end_param,
                    self.major_axis,
                    old_minor_axis,
                    self.center,
                    self.ratio,
                )
            )

            start_param = param(start_point - new_center)
            end_param = param(end_point - new_center)

            # Test if drawing the correct side of the curve
            if not math.isclose(old_param_span, math.pi, abs_tol=1e-9):
                # Equal param span check works well, except for a span of exact
                # pi (180 deg).
                # TODO: use ellipse_param_span()?
                #  2021-01-28 this is possibly the source of errors!
                new_param_span = (end_param - start_param) % math.tau
                if not math.isclose(
                    old_param_span, new_param_span, abs_tol=1e-9
                ):
                    start_param, end_param = end_param, start_param
            else:  # param span is exact pi (180 deg)
                # expensive but it seem to work:
                old_chk_point = m.transform(
                    vertex(
                        mid_param(old_start_param, old_end_param),
                        self.major_axis,
                        old_minor_axis,
                        self.center,
                        self.ratio,
                    )
                )
                new_chk_point = vertex(
                    mid_param(start_param, end_param),
                    new_major_axis,
                    new_minor_axis,
                    new_center,
                    new_ratio,
                )
                if not old_chk_point.isclose(new_chk_point, abs_tol=1e-9):
                    start_param, end_param = end_param, start_param

        if new_ratio > 1:
            new_major_axis = minor_axis(
                new_major_axis, new_extrusion, new_ratio
            )
            new_ratio = 1.0 / new_ratio
            new_minor_axis = minor_axis(
                new_major_axis, new_extrusion, new_ratio
            )
            if not (
                math.isclose(start_param, 0)
                and math.isclose(end_param, math.tau)
            ):
                start_param -= HALF_PI
                end_param -= HALF_PI

        # TODO: remove normalize start- and end params?
        #  2021-01-28 this is possibly the source of errors!
        start_param = start_param % math.tau
        end_param = end_param % math.tau
        if math.isclose(start_param, end_param):
            start_param = 0.0
            end_param = math.tau

        self.center = new_center
        self.major_axis = new_major_axis
        self.minor_axis = new_minor_axis
        self.extrusion = new_extrusion
        self.ratio = new_ratio
        self.start_param = start_param
        self.end_param = end_param

    @property
    def param_span(self) -> float:
        """Returns the counter clockwise params span from start- to end param,
        see also :func:`ezdxf.math.ellipse_param_span` for more information.

        """
        return arc_angle_span_rad(self.start_param, self.end_param)

    def params(self, num: int) -> Iterable[float]:
        """Returns `num` params from start- to end param in counter clockwise
        order.

        All params are normalized in the range from [0, 2π).

        """
        yield from get_params(self.start_param, self.end_param, num)

    def vertices(self, params: Iterable[float]) -> Iterable[Vec3]:
        """Yields vertices on ellipse for iterable `params` in WCS.

        Args:
            params: param values in the range from [0, 2π) in radians,
                param goes counter clockwise around the extrusion vector,
                major_axis = local x-axis = 0 rad.

        """
        center = self.center
        ratio = self.ratio
        x_axis = self.major_axis.normalize()
        y_axis = self.minor_axis.normalize()
        radius_x = self.major_axis.magnitude
        radius_y = radius_x * ratio

        for param in params:
            x = math.cos(param) * radius_x * x_axis
            y = math.sin(param) * radius_y * y_axis
            yield center + x + y

    def flattening(self, distance: float, segments: int = 4) -> Iterable[Vec3]:
        """Adaptive recursive flattening. The argument `segments` is the
        minimum count of approximation segments, if the distance from the center
        of the approximation segment to the curve is bigger than `distance` the
        segment will be subdivided. Returns a closed polygon for a full ellipse:
        start vertex == end vertex.

        Args:
            distance: maximum distance from the projected curve point onto the
                segment chord.
            segments: minimum segment count

        .. versionadded:: 0.15

        """

        def vertex_(p: float) -> Vec3:
            x = math.cos(p) * radius_x * x_axis
            y = math.sin(p) * radius_y * y_axis
            return self.center + x + y

        def subdiv(s: Vec3, e: Vec3, s_param: float, e_param: float):
            m_param = (s_param + e_param) * 0.5
            m = vertex_(m_param)
            d = distance_point_line_3d(m, s, e)
            if d < distance:
                yield e
            else:
                yield from subdiv(s, m, s_param, m_param)
                yield from subdiv(m, e, m_param, e_param)

        x_axis = self.major_axis.normalize()
        y_axis = self.minor_axis.normalize()
        radius_x = self.major_axis.magnitude
        radius_y = radius_x * self.ratio

        delta = self.param_span / segments
        if delta == 0.0:
            return

        param = self.start_param % math.tau
        if math.isclose(self.end_param, math.tau):
            end_param = math.tau
        else:
            end_param = self.end_param % math.tau

        if math.isclose(param, end_param):
            return
        elif param > end_param:
            end_param += math.tau

        start_point = vertex_(param)
        yield start_point
        while param < end_param:
            next_end_param = param + delta
            if math.isclose(next_end_param, end_param):
                next_end_param = end_param
            end_point = vertex_(next_end_param)
            yield from subdiv(start_point, end_point, param, next_end_param)
            param = next_end_param
            start_point = end_point

    def params_from_vertices(
        self, vertices: Iterable["Vertex"]
    ) -> Iterable[float]:
        """Yields ellipse params for all given `vertices`.

        The vertex don't has to be exact on the ellipse curve or in the range
        from start- to end param or even in the ellipse plane. Param is
        calculated from the intersection point of the ray projected on the
        ellipse plane from the center of the ellipse through the vertex.

        .. warning::

            An input for start- and end vertex at param 0 and 2π return
            unpredictable results because of floating point inaccuracy,
            sometimes 0 and sometimes 2π.

        """
        x_axis = self.major_axis.normalize()
        y_axis = self.minor_axis.normalize()
        ratio = self.ratio
        center = self.center
        for v in Vec3.generate(vertices):
            v -= center
            yield math.atan2(y_axis.dot(v) / ratio, x_axis.dot(v)) % math.tau

    def tangents(self, params: Iterable[float]) -> Iterable[Vec3]:
        """Yields tangents on ellipse for iterable `params` in WCS as direction
        vectors.

        Args:
            params: param values in the range from [0, 2π] in radians, param
                goes counter clockwise around the extrusion vector,
                major_axis = local x-axis = 0 rad.

        """
        ratio = self.ratio
        x_axis = self.major_axis.normalize()
        y_axis = self.minor_axis.normalize()

        for param in params:
            x = -math.sin(param) * x_axis
            y = math.cos(param) * ratio * y_axis
            yield (x + y).normalize()

    def swap_axis(self) -> None:
        """Swap axis and adjust start- and end parameter."""
        self.major_axis = self.minor_axis
        ratio = 1.0 / self.ratio
        self.ratio = max(ratio, 1e-6)
        self.minor_axis = minor_axis(
            self.major_axis, self.extrusion, self.ratio
        )

        start_param = self.start_param
        end_param = self.end_param
        if math.isclose(start_param, 0) and math.isclose(end_param, math.tau):
            return
        self.start_param = (start_param - HALF_PI) % math.tau
        self.end_param = (end_param - HALF_PI) % math.tau

    def add_to_layout(
        self, layout: "BaseLayout", dxfattribs=None
    ) -> "Ellipse":
        """Add ellipse as DXF :class:`~ezdxf.entities.Ellipse` entity to a
        layout.

        Args:
            layout: destination layout as :class:`~ezdxf.layouts.BaseLayout`
                object
            dxfattribs: additional DXF attributes for the ELLIPSE entity

        """
        from ezdxf.entities import Ellipse

        dxfattribs = dict(dxfattribs or {})
        dxfattribs.update(self.dxfattribs())
        e = Ellipse.new(dxfattribs=dxfattribs, doc=layout.doc)
        layout.add_entity(e)
        return e

    def to_ocs(self) -> "ConstructionEllipse":
        """Returns ellipse parameters as OCS representation.

        OCS elevation is stored in :attr:`center.z`.

        """
        ocs = OCS(self.extrusion)
        return self.__class__(
            center=ocs.from_wcs(self.center),
            major_axis=ocs.from_wcs(self.major_axis).replace(z=0),  # type: ignore
            ratio=self.ratio,
            start_param=self.start_param,
            end_param=self.end_param,
        )


def mid_param(start: float, end: float) -> float:
    if end < start:
        end += math.tau
    return (start + end) / 2.0


def minor_axis(major_axis: Vec3, extrusion: Vec3, ratio: float) -> Vec3:
    return extrusion.cross(major_axis).normalize(major_axis.magnitude * ratio)


def vertex(
    param: float, major_axis: Vec3, minor_axis: Vec3, center: Vec3, ratio: float
) -> Vec3:
    x_axis = major_axis.normalize()
    y_axis = minor_axis.normalize()
    radius_x = major_axis.magnitude
    radius_y = radius_x * ratio
    x = math.cos(param) * radius_x * x_axis
    y = math.sin(param) * radius_y * y_axis
    return center + x + y


def get_params(start: float, end: float, num: int) -> Iterable[float]:
    """Returns `num` params from start- to end param in counter clockwise order.

    All params are normalized in the range from [0, 2π).

    """
    if num < 2:
        raise ValueError("num >= 2")
    if end <= start:
        end += math.tau

    for param in linspace(start, end, num):
        yield param % math.tau


def angle_to_param(ratio: float, angle: float) -> float:
    """Returns ellipse parameter for argument `angle`.

    Args:
        ratio: minor axis to major axis ratio as stored in the ELLIPSE entity
            (always <= 1).
        angle: angle between major axis and line from center to point on the
            ellipse

    Returns:
        the ellipse parameter in the range [0, 2π)
    """
    return math.atan2(math.sin(angle) / ratio, math.cos(angle)) % math.tau


def param_to_angle(ratio: float, param: float) -> float:
    """Returns circle angle from ellipse parameter for argument `angle`.

    Args:
        ratio: minor axis to major axis ratio as stored in the ELLIPSE entity
            (always <= 1).
        param: ellipse parameter between major axis and point on the ellipse
            curve

    Returns:
        the circle angle in the range [0, 2π)
    """
    return math.atan2(math.sin(param) * ratio, math.cos(param))


def rytz_axis_construction(d1: Vec3, d2: Vec3) -> Tuple[Vec3, Vec3, float]:
    """The Rytz’s axis construction is a basic method of descriptive Geometry
    to find the axes, the semi-major axis and semi-minor axis, starting from two
    conjugated half-diameters.

    Source: `Wikipedia <https://en.m.wikipedia.org/wiki/Rytz%27s_construction>`_

    Given conjugated diameter `d1` is the vector from center C to point P and
    the given conjugated diameter `d2` is the vector from center C to point Q.
    Center of ellipse is always ``(0, 0, 0)``. This algorithm works for
    2D/3D vectors.

    Args:
        d1: conjugated semi-major axis as :class:`Vec3`
        d2: conjugated semi-minor axis as :class:`Vec3`

    Returns:
         Tuple of (major axis, minor axis, ratio)

    """
    Q = Vec3(d1)  # vector CQ
    # calculate vector CP', location P'
    if math.isclose(d1.z, 0, abs_tol=1e-9) and math.isclose(
        d2.z, 0, abs_tol=1e-9
    ):
        # Vec3.orthogonal() works only for vectors in the xy-plane!
        P1 = Vec3(d2).orthogonal(ccw=False)
    else:
        extrusion = d1.cross(d2)
        P1 = extrusion.cross(d2).normalize(d2.magnitude)

    D = P1.lerp(Q)  # vector CD, location D, midpoint of P'Q
    radius = D.magnitude
    radius_vector = (Q - P1).normalize(radius)  # direction vector P'Q
    A = D - radius_vector  # vector CA, location A
    B = D + radius_vector  # vector CB, location B
    if A.isclose(NULLVEC) or B.isclose(NULLVEC):
        raise ArithmeticError("Conjugated axis required, invalid source data.")
    major_axis_length = (A - Q).magnitude
    minor_axis_length = (B - Q).magnitude
    if math.isclose(major_axis_length, 0.0) or math.isclose(
        minor_axis_length, 0.0
    ):
        raise ArithmeticError("Conjugated axis required, invalid source data.")
    ratio = minor_axis_length / major_axis_length
    major_axis = B.normalize(major_axis_length)
    minor_axis = A.normalize(minor_axis_length)
    return major_axis, minor_axis, ratio



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import Dict, Iterable, Sequence
from .mtext import MText, MTextColumns

__all__ = [
    "make_static_columns_r2000",
    "make_dynamic_auto_height_columns_r2000",
    "make_dynamic_manual_height_columns_r2000",
    "make_static_columns_r2018",
    "make_dynamic_auto_height_columns_r2018",
    "make_dynamic_manual_height_columns_r2018",
]

COLUMN_BREAK = "\\N"


def add_column_breaks(content: Iterable[str]) -> Iterable[str]:
    content = list(content)
    for c in content[:-1]:
        if not c.endswith(COLUMN_BREAK):
            c += COLUMN_BREAK
        yield c
    yield content[-1]  # last column without a column break


def make_static_columns_r2000(
        content: Sequence[str],
        width: float, gutter_width: float, height: float,
        dxfattribs: Dict = None) -> MText:
    if len(content) < 1:
        raise ValueError("no content")
    columns = MTextColumns.new_static_columns(
        len(content), width, gutter_width, height)
    mtext = MText.new(dxfattribs=dxfattribs)
    mtext.setup_columns(columns, linked=True)
    content = list(add_column_breaks(content))
    mtext.text = content[0]
    for mt, c in zip(columns.linked_columns, content[1:]):
        mt.text = c
    return mtext


def make_dynamic_auto_height_columns_r2000(
        content: str,
        width: float, gutter_width: float, height: float, count: int = 1,
        dxfattribs: Dict = None) -> MText:
    if not content:
        raise ValueError("no content")
    mtext = MText.new(dxfattribs=dxfattribs)
    mtext.dxf.width = width
    columns = MTextColumns.new_dynamic_auto_height_columns(
        count, width, gutter_width, height)
    set_dynamic_columns_content(content, mtext, columns)
    return mtext


def make_dynamic_manual_height_columns_r2000(
        content: str,
        width: float, gutter_width: float, heights: Sequence[float],
        dxfattribs: Dict = None) -> MText:
    if not content:
        raise ValueError("no content")
    mtext = MText.new(dxfattribs=dxfattribs)
    mtext.dxf.width = width
    columns = MTextColumns.new_dynamic_manual_height_columns(
        width, gutter_width, heights)
    set_dynamic_columns_content(content, mtext, columns)
    return mtext


def set_dynamic_columns_content(
        content: str, mtext: MText, columns: MTextColumns):
    mtext.setup_columns(columns, linked=True)
    # temp. hack: assign whole content to the main column
    mtext.text = content
    # for mt, c in zip(mtext.columns.linked_columns, content[1:]):
    #    mt.text = c
    return mtext


# DXF version R2018

def make_static_columns_r2018(
        content: Sequence[str],
        width: float, gutter_width: float, height: float,
        dxfattribs: Dict = None) -> MText:
    if len(content) < 1:
        raise ValueError("no content")
    columns = MTextColumns.new_static_columns(
        len(content), width, gutter_width, height)
    mtext = MText.new(dxfattribs=dxfattribs)
    mtext.setup_columns(columns, linked=False)
    mtext.text = ''.join(add_column_breaks(content))
    return mtext


def make_dynamic_auto_height_columns_r2018(
        content: str,
        width: float, gutter_width: float, height: float, count: int,
        dxfattribs: Dict = None) -> MText:
    columns = MTextColumns.new_dynamic_auto_height_columns(
        count, width, gutter_width, height)
    return _make_dynamic_columns_r2018(content, columns, dxfattribs or {})


def make_dynamic_manual_height_columns_r2018(
        content: str,
        width: float, gutter_width: float, heights: Sequence[float],
        dxfattribs: Dict = None) -> MText:
    columns = MTextColumns.new_dynamic_manual_height_columns(
        width, gutter_width, heights)
    return _make_dynamic_columns_r2018(content, columns, dxfattribs or {})


def _make_dynamic_columns_r2018(
        content: str, columns: MTextColumns, dxfattribs: Dict) -> MText:
    if not content:
        raise ValueError("no content")
        # column count is not required for DXF R2018
    mtext = MText.new(dxfattribs=dxfattribs)
    mtext.setup_columns(columns, linked=False)
    mtext.text = content
    return mtext



# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterator
from ezdxf.entities import factory, is_graphic_entity
from ezdxf.enums import InsertUnits
from ezdxf.lldxf.const import (
    DXFValueError,
    DXFStructureError,
    LATEST_DXF_VERSION,
    DXFTypeError,
)
from ezdxf.query import EntityQuery
from ezdxf.groupby import groupby
from ezdxf.entitydb import EntityDB, EntitySpace
from ezdxf.graphicsfactory import CreatorInterface

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        BlockRecord,
        DXFGraphic,
        KeyFunc,
        ExtensionDict,
    )

SUPPORTED_FOREIGN_ENTITY_TYPES = {
    "ARC",
    "LINE",
    "CIRCLE",
    "ELLIPSE",
    "POINT",
    "LWPOLYLINE",
    "SPLINE",
    "3DFACE",
    "SOLID",
    "TRACE",
    "SHAPE",
    "POLYLINE",
    "MESH",
    "TEXT",
    "MTEXT",
    "HATCH",
    "ATTRIB",
    "ATTDEF",
}


class _AbstractLayout(CreatorInterface):
    entity_space = EntitySpace()

    @property
    def entitydb(self) -> "EntityDB":
        """Returns drawing entity database. (internal API)"""
        return self.doc.entitydb

    def rename(self, name) -> None:
        pass

    def __len__(self) -> int:
        """Returns count of entities owned by the layout."""
        return len(self.entity_space)

    def __iter__(self) -> Iterator["DXFGraphic"]:
        """Returns iterable of all drawing entities in this layout."""
        return iter(self.entity_space)  # type: ignore

    def __getitem__(self, index):
        """Get entity at `index`.

        The underlying data structure for storing entities is organized like a
        standard Python list, therefore `index` can be any valid list indexing
        or slicing term, like a single index ``layout[-1]`` to get the last
        entity, or an index slice ``layout[:10]`` to get the first 10 or less
        entities as ``List[DXFGraphic]``.

        """
        return self.entity_space[index]

    def query(self, query: str = "*") -> EntityQuery:
        """Get all DXF entities matching the :ref:`entity query string`."""
        return EntityQuery(iter(self), query)

    def groupby(self, dxfattrib: str = "", key: "KeyFunc" = None) -> dict:
        """
        Returns a ``dict`` of entity lists, where entities are grouped by a
        `dxfattrib` or a `key` function.

        Args:
            dxfattrib: grouping by DXF attribute like ``'layer'``
            key: key function, which accepts a :class:`DXFGraphic` entity as
                argument and returns the grouping key of an entity or ``None``
                to ignore the entity. Reason for ignoring: a queried DXF
                attribute is not supported by entity.

        """
        return groupby(iter(self), dxfattrib, key)

    def destroy(self):
        pass

    def purge(self):
        """Remove all destroyed entities from the layout entity space."""
        self.entity_space.purge()


class BaseLayout(_AbstractLayout):
    def __init__(self, block_record: "BlockRecord"):
        doc = block_record.doc
        assert doc is not None
        super().__init__(doc)
        self.entity_space = block_record.entity_space
        # This is the real central layout management structure:
        self.block_record: "BlockRecord" = block_record

    @property
    def block_record_handle(self):
        """Returns block record handle. (internal API)"""
        return self.block_record.dxf.handle

    @property
    def layout_key(self) -> str:
        """Returns the layout key as hex string.

        The layout key is the handle of the associated BLOCK_RECORD entry in the
        BLOCK_RECORDS table.

        (internal API)
        """
        return self.block_record.dxf.handle

    @property
    def is_alive(self):
        """``False`` if layout is deleted."""
        return self.block_record.is_alive

    @property
    def is_active_paperspace(self) -> bool:
        """``True`` if is active layout."""
        return self.block_record.is_active_paperspace

    @property
    def is_any_paperspace(self) -> bool:
        """``True`` if is any kind of paperspace layout."""
        return self.block_record.is_any_paperspace

    @property
    def is_modelspace(self) -> bool:
        """``True`` if is modelspace layout."""
        return self.block_record.is_modelspace

    @property
    def is_any_layout(self) -> bool:
        """``True`` if is any kind of modelspace or paperspace layout."""
        return self.block_record.is_any_layout

    @property
    def is_block_layout(self) -> bool:
        """``True`` if not any kind of modelspace or paperspace layout, just a
        regular block definition.
        """
        return self.block_record.is_block_layout

    @property
    def units(self) -> InsertUnits:
        """Get/Set layout/block drawing units as enum, see also :ref:`set
        drawing units`.
        """
        # todo: doesn't care about units stored in XDATA, see ezdxf/units.py
        # Don't know what this units are used for, but header var $INSUNITS
        # are the real units of the model space.
        return self.block_record.dxf.units

    @units.setter
    def units(self, value: InsertUnits) -> None:
        """Set layout/block drawing units as enum."""
        self.block_record.dxf.units = value  # has a DXF validator

    def get_extension_dict(self) -> "ExtensionDict":
        """Returns the associated extension dictionary, creates a new one if
        necessary.
        """
        block_record = self.block_record
        if block_record.has_extension_dict:
            return block_record.get_extension_dict()
        else:
            return block_record.new_extension_dict()

    def add_entity(self, entity: "DXFGraphic") -> None:
        """Add an existing :class:`DXFGraphic` entity to a layout, but be sure
        to unlink (:meth:`~BaseLayout.unlink_entity`) entity from the previous
        owner layout. Adding entities from a different DXF drawing is not
        supported.
        """
        # bind virtual entities to the DXF document:
        doc = self.doc
        if entity.dxf.handle is None and doc:
            factory.bind(entity, doc)
        handle = entity.dxf.handle
        if handle is None or handle not in self.doc.entitydb:
            raise DXFStructureError(
                "Adding entities from a different DXF drawing is not supported."
            )

        if not is_graphic_entity(entity):
            raise DXFTypeError(f"invalid entity {str(entity)}")
        self.block_record.add_entity(entity)

    def add_foreign_entity(self, entity: "DXFGraphic", copy=True) -> None:
        """Add a foreign DXF entity to a layout, this foreign entity could be
        from another DXF document or an entity without an assigned DXF document.
        The intention of this method is to add **simple** entities from another
        DXF document or from a DXF iterator, for more complex operations use the
        :mod:`~ezdxf.addons.importer` add-on. Especially objects with BLOCK
        section (INSERT, DIMENSION, MLEADER) or OBJECTS section dependencies
        (IMAGE, UNDERLAY) can not be supported by this simple method.

        Not all DXF types are supported and every dependency or resource
        reference from another DXF document will be removed except attribute
        layer will be preserved but only with default attributes like
        color ``7`` and linetype ``CONTINUOUS`` because the layer attribute
        doesn't need a layer table entry.

        If the entity is part of another DXF document, it will be unlinked from
        this document and its entity database if argument `copy` is ``False``,
        else the entity will be copied. Unassigned entities like from DXF
        iterators will just be added.

        Supported DXF types:

            - POINT
            - LINE
            - CIRCLE
            - ARC
            - ELLIPSE
            - LWPOLYLINE
            - SPLINE
            - POLYLINE
            - 3DFACE
            - SOLID
            - TRACE
            - SHAPE
            - MESH
            - ATTRIB
            - ATTDEF
            - TEXT
            - MTEXT
            - HATCH

        Args:
            entity: DXF entity to copy or move
            copy: if ``True`` copy entity from other document else unlink from
                other document

        """
        foreign_doc = entity.doc
        dxftype = entity.dxftype()
        if dxftype not in SUPPORTED_FOREIGN_ENTITY_TYPES:
            raise DXFTypeError(f"unsupported DXF type: {dxftype}")
        if foreign_doc is self.doc:
            raise DXFValueError("entity from same DXF document")

        if foreign_doc is not None:
            if copy:
                entity = entity.copy()
            else:
                # Unbind entity from other document without destruction.
                factory.unbind(entity)

        entity.remove_dependencies(self.doc)
        # add to this layout & bind to document
        self.add_entity(entity)

    def unlink_entity(self, entity: "DXFGraphic") -> None:
        """Unlink `entity` from layout but does not delete entity from the
        entity database, this removes `entity` just from the layout entity space.
        """
        self.block_record.unlink_entity(entity)

    def delete_entity(self, entity: "DXFGraphic") -> None:
        """Delete `entity` from layout entity space and the entity database,
        this destroys the `entity`.
        """
        self.block_record.delete_entity(entity)

    def delete_all_entities(self) -> None:
        """Delete all entities from this layout and from entity database,
        this destroys all entities in this layout.
        """
        # Create list, because delete modifies the base data structure of
        # the iterator:
        for entity in list(self):
            self.delete_entity(entity)

    def move_to_layout(
        self, entity: "DXFGraphic", layout: "BaseLayout"
    ) -> None:
        """Move entity to another layout.

        Args:
            entity: DXF entity to move
            layout: any layout (modelspace, paperspace, block) from
                **same** drawing

        """
        if entity.doc != layout.doc:
            raise DXFStructureError(
                "Moving between different DXF drawings is not supported."
            )
        try:
            self.unlink_entity(entity)
        except ValueError:
            raise DXFValueError("Layout does not contain entity.")
        else:
            layout.add_entity(entity)

    def destroy(self) -> None:
        """Delete all linked resources. (internal API)"""
        # block_records table is owner of block_record has to delete it
        # the block_record is the owner of the entities and deletes them all
        self.doc.block_records.remove(self.block_record.dxf.name)


class VirtualLayout(_AbstractLayout):
    """Helper class to disassemble complex entities into basic DXF
    entities by rendering into a virtual layout.

    All entities do not have an assigned DXF document and therefore
    are not stored in any entity database and can not be added to another
    layout by :meth:`add_entity`.

    Deleting entities from this layout does not destroy the entity!

    """

    def __init__(self):
        super().__init__(None)
        self.entity_space = EntitySpace()

    @property
    def dxfversion(self) -> str:
        return LATEST_DXF_VERSION

    def add_entity(self, entity: "DXFGraphic") -> None:
        self.entity_space.add(entity)

    def new_entity(self, type_: str, dxfattribs: dict) -> "DXFGraphic":
        entity = factory.new(type_, dxfattribs=dxfattribs)
        self.entity_space.add(entity)
        return entity  # type: ignore

    def unlink_entity(self, entity: "DXFGraphic") -> None:
        self.entity_space.remove(entity)

    def delete_entity(self, entity: "DXFGraphic") -> None:
        self.entity_space.remove(entity)

    def delete_all_entities(self) -> None:
        self.entity_space.clear()

    def copy_all_to_layout(self, layout: BaseLayout) -> None:
        """Copy all entities to a real document layout."""
        doc = layout.doc
        entitydb = doc.entitydb
        for entity in self.entity_space:
            try:
                clone = entity.copy()
            except DXFTypeError:
                continue
            clone.doc = doc
            entitydb.add(clone)
            layout.add_entity(clone)  # type: ignore

    def move_all_to_layout(self, layout: BaseLayout) -> None:
        """Move all entities to a real document layout."""
        doc = layout.doc
        entitydb = doc.entitydb
        for entity in self.entity_space:
            entity.doc = doc
            entitydb.add(entity)
            layout.add_entity(entity)  # type: ignore
        self.delete_all_entities()



# Copyright (c) 2010-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, List, Union, Tuple

from functools import partial
import math
import warnings
from ezdxf.math import Vec3, Vec2, Matrix44, X_AXIS, Y_AXIS, arc_angle_span_rad
from decimal import Decimal

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex

TOLERANCE = 1e-10
RADIANS_90 = math.pi / 2.0
RADIANS_180 = math.pi
RADIANS_270 = RADIANS_90 * 3.0
RADIANS_360 = 2.0 * math.pi

__all__ = [
    "is_close_points",
    "closest_point",
    "convex_hull_2d",
    "distance_point_line_2d",
    "is_point_on_line_2d",
    "is_point_in_polygon_2d",
    "is_point_left_of_line",
    "point_to_line_relation",
    "linspace",
    "enclosing_angles",
    "reflect_angle_x_deg",
    "reflect_angle_y_deg",
    "sign",
    "area",
    "circle_radius_3p",
    "TOLERANCE",
    "has_matrix_2d_stretching",
    "decdeg2dms",
    "ellipse_param_span",
]


def is_close_points(p1: "Vertex", p2: "Vertex", abs_tol=TOLERANCE) -> bool:
    """Deprecated function will be removed in v0.18! Use Vec(p1).isclose(p2)."""
    warnings.warn(
        "Deprecated function will be removed in v0.18! "
        "Use Vec3(p1).isclose(p2).",
        DeprecationWarning,
    )
    return Vec3(p1).isclose(p2, abs_tol=abs_tol)


def linspace(
    start: float, stop: float, num: int, endpoint=True
) -> Iterable[float]:
    """Return evenly spaced numbers over a specified interval, like
    numpy.linspace().

    Returns `num` evenly spaced samples, calculated over the interval
    [start, stop]. The endpoint of the interval can optionally be excluded.

    """
    if num < 0:
        raise ValueError(f"Number of samples, {num}, must be non-negative.")
    elif num == 0:
        return
    elif num == 1:
        yield start
        return

    start_dec = Decimal(start)
    count = (num - 1) if endpoint else num
    delta = (Decimal(stop) - start_dec) / count
    for _ in range(num):
        yield float(start_dec)
        start_dec += delta


def sign(f: float) -> float:
    """Return sign of float `f` as -1 or +1, 0 returns +1"""
    return -1.0 if f < 0.0 else +1.0


def reflect_angle_x_deg(a: float) -> float:
    """Returns reflected angle of `a` in x-direction in degrees.
    Angles are counter clockwise orientated and +x-axis is at 0 degrees.

    Args:
        a: angle to reflect in degrees

    """
    return (180.0 - (a % 360.0)) % 360.0


def reflect_angle_y_deg(a: float) -> float:
    """Returns reflected angle of `a` in y-direction in degrees.
    Angles are counter clockwise orientated and +y-axis is at 90 degrees.

    Args:
        a: angle to reflect in degrees

    """
    return (360.0 - (a % 360.0)) % 360.0


def decdeg2dms(value: float) -> Tuple[float, float, float]:
    """Return decimal degrees as tuple (Degrees, Minutes, Seconds)."""
    mnt, sec = divmod(value * 3600., 60.)
    deg, mnt = divmod(mnt, 60.)
    return deg, mnt, sec


def ellipse_param_span(start_param: float, end_param: float) -> float:
    """Returns the counter clockwise params span of an elliptic arc from start-
    to end param.

    Returns the param span in the range [0, 2π], 2π is a full ellipse.
    Full ellipse handling is a special case, because normalization of params
    which describe a full ellipse would return 0 if treated as regular params.
    e.g. (0, 2π) → 2π, (0, -2π) → 2π, (π, -π) → 2π.
    Input params with the same value always return 0 by definition:
    (0, 0) → 0, (-π, -π) → 0, (2π, 2π) → 0.

    Alias to function: :func:`ezdxf.math.arc_angle_span_rad`

    """
    return arc_angle_span_rad(float(start_param), float(end_param))


def closest_point(base: "Vertex", points: Iterable["Vertex"]) -> "Vec3":
    """Returns closest point to `base`.

    Args:
        base: base point as :class:`Vec3` compatible object
        points: iterable of points as :class:`Vec3` compatible object

    """
    base = Vec3(base)
    min_dist = None
    found = None
    for point in points:
        p = Vec3(point)
        dist = (base - p).magnitude
        if (min_dist is None) or (dist < min_dist):
            min_dist = dist
            found = p
    return found


def convex_hull_2d(points: Iterable["Vertex"]) -> List["Vertex"]:
    """Returns 2D convex hull for `points`.

    Args:
        points: iterable of points as :class:`Vec3` compatible objects,
            z-axis is ignored

    """

    def _convexhull(hull):
        while len(hull) > 2:
            # the last three points
            start_point, check_point, destination_point = hull[-3:]
            # curve not turns right
            if not is_point_left_of_line(
                check_point, start_point, destination_point
            ):
                # remove the penultimate point
                del hull[-2]
            else:
                break
        return hull

    points = sorted(set(Vec2.generate(points)))  # remove duplicate points

    if len(points) < 3:
        raise ValueError(
            "Convex hull calculation requires 3 or more unique points."
        )

    upper_hull = points[:2]  # first two points
    for next_point in points[2:]:
        upper_hull.append(next_point)
        upper_hull = _convexhull(upper_hull)
    lower_hull = [points[-1], points[-2]]  # last two points

    for next_point in reversed(points[:-2]):
        lower_hull.append(next_point)
        lower_hull = _convexhull(lower_hull)
    upper_hull.extend(lower_hull[1:-1])
    return upper_hull


def enclosing_angles(
    angle, start_angle, end_angle, ccw=True, abs_tol=TOLERANCE
):
    isclose = partial(math.isclose, abs_tol=abs_tol)

    s = start_angle % math.tau
    e = end_angle % math.tau
    a = angle % math.tau
    if isclose(s, e):
        return isclose(s, a)

    if s < e:
        r = s < a < e
    else:
        r = not (e < a < s)
    return r if ccw else not r


def is_point_on_line_2d(
    point: Vec2, start: Vec2, end: Vec2, ray=True, abs_tol=TOLERANCE
) -> bool:
    """Returns ``True`` if `point` is on `line`.

    Args:
        point: 2D point to test as :class:`Vec2`
        start: line definition point as :class:`Vec2`
        end: line definition point as :class:`Vec2`
        ray: if ``True`` point has to be on the infinite ray, if ``False``
            point has to be on the line segment
        abs_tol: tolerance for on line test

    """
    point_x, point_y = point
    start_x, start_y = start
    end_x, end_y = end
    on_line = (
        math.fabs(
            (end_y - start_y) * point_x
            - (end_x - start_x) * point_y
            + (end_x * start_y - end_y * start_x)
        )
        <= abs_tol
    )
    if not on_line or ray:
        return on_line
    else:
        if start_x > end_x:
            start_x, end_x = end_x, start_x
        if not (start_x - abs_tol <= point_x <= end_x + abs_tol):
            return False
        if start_y > end_y:
            start_y, end_y = end_y, start_y
        if not (start_y - abs_tol <= point_y <= end_y + abs_tol):
            return False
        return True


def point_to_line_relation(
    point: Vec2, start: Vec2, end: Vec2, abs_tol=TOLERANCE
) -> int:
    """Returns ``-1`` if `point` is left `line`, ``+1`` if `point` is right of
    `line` and ``0`` if `point` is on the `line`. The `line` is defined by two
    vertices given as arguments `start` and `end`.

    Args:
        point: 2D point to test as :class:`Vec2`
        start: line definition point as :class:`Vec2`
        end: line definition point as :class:`Vec2`
        abs_tol: tolerance for minimum distance to line

    """
    rel = (end.x - start.x) * (point.y - start.y) - (end.y - start.y) * (
        point.x - start.x
    )
    if abs(rel) <= abs_tol:
        return 0
    elif rel < 0:
        return +1
    else:
        return -1


def is_point_left_of_line(
    point: Vec2, start: Vec2, end: Vec2, colinear=False
) -> bool:
    """Returns ``True`` if `point` is "left of line" defined by `start-` and
    `end` point, a colinear point is also "left of line" if argument `colinear`
    is ``True``.

    Args:
        point: 2D point to test as :class:`Vec2`
        start: line definition point as :class:`Vec2`
        end: line definition point as :class:`Vec2`
        colinear: a colinear point is also "left of line" if ``True``

    """
    rel = point_to_line_relation(point, start, end)
    if colinear:
        return rel < 1
    else:
        return rel < 0


def distance_point_line_2d(point: Vec2, start: Vec2, end: Vec2) -> float:
    """Returns the normal distance from `point` to 2D line defined by `start-`
    and `end` point.
    """
    # wikipedia: https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line.
    if start.isclose(end):
        raise ZeroDivisionError("Not a line.")
    return math.fabs((start - point).det(end - point)) / (end - start).magnitude


def is_point_in_polygon_2d(
    point: Union[Vec2, Vec3], polygon: Iterable[Vec2], abs_tol=TOLERANCE
) -> int:
    """Test if `point` is inside `polygon`.

    Args:
        point: 2D point to test as :class:`Vec2`
        polygon: iterable of 2D points as :class:`Vec2`
        abs_tol: tolerance for distance check

    Returns:
        ``+1`` for inside, ``0`` for on boundary line, ``-1`` for outside

    """
    # Source: http://www.faqs.org/faqs/graphics/algorithms-faq/
    # Subject 2.03: How do I find if a point lies within a polygon?
    polygon = list(polygon)  # shallow copy, because list will be modified
    if not polygon[0].isclose(polygon[-1]):
        polygon.append(polygon[0])
    if len(polygon) < 4:  # 3+1 because first point == last point
        raise ValueError("At least 3 polygon points required.")
    x = point.x
    y = point.y
    # ignore z-axis of Vec3()
    inside = False
    for i in range(len(polygon) - 1):
        x1, y1 = polygon[i]
        x2, y2 = polygon[i + 1]
        # is point on polygon boundary line:
        # is point in x-range of line
        a, b = (x2, x1) if x2 < x1 else (x1, x2)
        if a <= x <= b:
            # is point in y-range of line
            c, d = (y2, y1) if y2 < y1 else (y1, y2)
            if (c <= y <= d) and math.fabs(
                (y2 - y1) * x - (x2 - x1) * y + (x2 * y1 - y2 * x1)
            ) <= abs_tol:
                return 0
        if ((y1 <= y < y2) or (y2 <= y < y1)) and (
            x < (x2 - x1) * (y - y1) / (y2 - y1) + x1
        ):
            inside = not inside
    if inside:
        return 1
    else:
        return -1


def circle_radius_3p(a: Vec3, b: Vec3, c: Vec3) -> float:
    ba = b - a
    ca = c - a
    cb = c - b
    upper = ba.magnitude * ca.magnitude * cb.magnitude
    lower = ba.cross(ca).magnitude * 2.0
    return upper / lower


def area(vertices: Iterable["Vertex"]) -> float:
    """Returns the area of a polygon, returns the projected area in the
    xy-plane for 3D vertices.
    """
    _vertices = Vec3.list(vertices)
    if len(_vertices) < 3:
        raise ValueError("At least 3 vertices required.")

    # Close polygon:
    if not _vertices[0].isclose(_vertices[-1]):
        _vertices.append(_vertices[0])

    return abs(
        sum(
            (p1.x * p2.y - p1.y * p2.x)
            for p1, p2 in zip(_vertices, _vertices[1:])
        )
        / 2
    )


def has_matrix_2d_stretching(m: Matrix44) -> bool:
    """Returns ``True`` if matrix `m` performs a non-uniform xy-scaling.
    Uniform scaling is not stretching in this context.

    Does not check if the target system is a cartesian coordinate system, use the
    :class:`~ezdxf.math.Matrix44` property :attr:`~ezdxf.math.Matrix44.is_cartesian`
    for that.
    """
    ux = m.transform_direction(X_AXIS)
    uy = m.transform_direction(Y_AXIS)
    return not math.isclose(ux.magnitude_square, uy.magnitude_square)



# Copyright (c) 2011-2021, Manfred Moitzi
# License: MIT License
from typing import TextIO, List, Union, Tuple
import os
import sys
from pathlib import Path
from configparser import ConfigParser

# Recommended uses of the global object "options":
# import ezdxf
# value = ezdxf.options.<attribute>
#
# If the import of "ezdxf" is not wanted:
# from ezdxf._options import options

# The MATPLOTLIB global shows that Matplotlib is installed:
try:
    import matplotlib

    MATPLOTLIB = True
except ImportError:
    MATPLOTLIB = False

TRUE_STATE = {"True", "true", "On", "on", "1"}
CORE = "core"
BROWSE_COMMAND = "browse-command"
VIEW_COMMAND = "view-command"
DRAW_COMMAND = "draw-command"
EZDXF_INI = "ezdxf.ini"
EZDXF = "ezdxf"
XDG_CONFIG_HOME = "XDG_CONFIG_HOME"
CONFIG_DIRECTORY = ".config"
ODAFC_ADDON = "odafc-addon"
DRAWING_ADDON = "drawing-addon"


def xdg_path(xdg_var: str, directory: str) -> Path:
    xdg_home = os.environ.get(xdg_var)
    if xdg_home:
        # should default to $HOME/<directory> e.g. $HOME/.config
        home = Path(xdg_home).expanduser()
    else:
        # replicate structure
        home = Path("~").expanduser() / directory
    return home / EZDXF


def config_home_path() -> Path:
    return xdg_path(XDG_CONFIG_HOME, CONFIG_DIRECTORY)


def default_config_files() -> List[Path]:
    config_paths = [
        config_home_path() / EZDXF_INI,
        Path(f"./{EZDXF_INI}"),
    ]
    return config_paths


def default_config() -> ConfigParser:
    config = ConfigParser()
    config[CORE] = {
        "DEFAULT_DIMENSION_TEXT_STYLE": "OpenSansCondensed-Light",
        "TEST_FILES": "",
        "FONT_CACHE_DIRECTORY": "",
        "LOAD_PROXY_GRAPHICS": "true",
        "STORE_PROXY_GRAPHICS": "true",
        "LOG_UNPROCESSED_TAGS": "false",
        "FILTER_INVALID_XDATA_GROUP_CODES": "true",
        "WRITE_FIXED_META_DATA_FOR_TESTING": "false",
        "DISABLE_C_EXT": "false",
    }
    config[BROWSE_COMMAND] = {
        "TEXT_EDITOR": r'"C:\Program Files\Notepad++\notepad++.exe" '
                       r'"{filename}" -n{num}',
        "ICON_SIZE": "32",
    }
    config[ODAFC_ADDON] = {
        "WIN_EXEC_PATH": r'"C:\Program Files\ODA\ODAFileConverter\ODAFileConverter.exe"'
    }
    config[DRAWING_ADDON] = {
        # These options are just for testing scenarios!
        "TRY_PYSIDE6": "true",
        "TRY_PYQT5": "true",
    }
    return config


def config_files() -> List[Path]:
    # Loading order for config files:
    # 1. user home directory:
    #    "$XDG_CONFIG_HOME/ezdxf/ezdxf.ini" or
    #    "~/.config/ezdxf/ezdxf.ini"
    # 2. current working directory "./ezdxf.ini"
    # 3. config file specified by EZDXF_CONFIG_FILE

    paths = default_config_files()
    env_cfg = os.getenv("EZDXF_CONFIG_FILE", "")
    if env_cfg:
        paths.append(Path(env_cfg))
    return paths


def load_config_files(paths: List[Path]) -> ConfigParser:
    config = default_config()
    config.read(paths, encoding="utf8")

    # environment variables override config files
    for name, env_name in [
        ("TEST_FILES", "EZDXF_TEST_FILES"),
        ("DISABLE_C_EXT", "EZDXF_DISABLE_C_EXT"),
    ]:
        value = os.environ.get(env_name, "")
        if value:
            config[CORE][name] = value
    return config


def boolstr(value: bool) -> str:
    return str(value).lower()


class Options:
    CORE = CORE
    BROWSE_COMMAND = BROWSE_COMMAND
    VIEW_COMMAND = VIEW_COMMAND
    DRAW_COMMAND = DRAW_COMMAND

    CONFIG_VARS = [
        "EZDXF_DISABLE_C_EXT",
        "EZDXF_TEST_FILES",
        "EZDXF_CONFIG_FILE",
    ]

    def __init__(self):
        paths = config_files()
        self._loaded_paths: List[Path] = [p for p in paths if p.exists()]
        self._config = load_config_files(paths)
        # needs fast access:
        self.log_unprocessed_tags = True
        # Activate/deactivate Matplotlib support (e.g. for testing)
        self._use_matplotlib = MATPLOTLIB
        self._use_c_ext = False  # set ezdxf.acc.__init__!
        self.update_cached_options()

    def set(self, section: str, key: str, value: str) -> None:
        self._config.set(section, key, value)

    def get(self, section: str, key: str, default: str = "") -> str:
        return self._config.get(section, key, fallback=default)

    def get_bool(self, section: str, key: str, default: bool = False) -> bool:
        return self._config.getboolean(section, key, fallback=default)

    def get_int(self, section: str, key: str, default: int = 0) -> int:
        return self._config.getint(section, key, fallback=default)

    def get_float(self, section: str, key: str, default: float = 0.0) -> float:
        return self._config.getfloat(section, key, fallback=default)

    def update_cached_options(self) -> None:
        self.log_unprocessed_tags = self.get_bool(
            Options.CORE, "LOG_UNPROCESSED_TAGS", default=True
        )

    def rewrite_cached_options(self):
        # rewrite cached options
        self._config.set(
            Options.CORE,
            "LOG_UNPROCESSED_TAGS",
            boolstr(self.log_unprocessed_tags),
        )

    @property
    def loaded_config_files(self) -> Tuple[Path, ...]:
        return tuple(self._loaded_paths)

    def read_file(self, filename: str) -> None:
        """Append content from config file `filename`, but does not reset the
        configuration.
        """
        try:
            self._config.read(filename)
        except IOError as e:
            print(str(e))
        else:
            self._loaded_paths.append(Path(filename))
            self.update_cached_options()

    def write(self, fp: TextIO) -> None:
        """Write current configuration into given file object, the file object
        must be a writeable text file with 'utf8' encoding.
        """
        self.rewrite_cached_options()
        try:
            self._config.write(fp)
        except IOError as e:
            print(str(e))

    def write_file(self, filename: str = EZDXF_INI) -> None:
        """Write current configuration into file `filename`."""
        with open(os.path.expanduser(filename), "wt", encoding="utf8") as fp:
            self.write(fp)

    @property
    def filter_invalid_xdata_group_codes(self) -> bool:
        return self.get_bool(
            CORE, "FILTER_INVALID_XDATA_GROUP_CODES", default=True
        )

    @property
    def default_dimension_text_style(self) -> str:
        return self.get(
            CORE,
            "DEFAULT_DIMENSION_TEXT_STYLE",
            default="OpenSansCondensed-Light",
        )

    @default_dimension_text_style.setter
    def default_dimension_text_style(self, style: str) -> None:
        self.set(
            CORE,
            "DEFAULT_DIMENSION_TEXT_STYLE",
            style,
        )

    @property
    def font_cache_directory(self) -> str:
        return os.path.expanduser(self.get(CORE, "FONT_CACHE_DIRECTORY"))

    @font_cache_directory.setter
    def font_cache_directory(self, dirname: Union[str, Path]) -> None:
        p = Path(dirname).expanduser()
        if p.exists():
            absolute = p.absolute()
            if p.is_dir():
                self.set(CORE, "FONT_CACHE_DIRECTORY", str(absolute))
            else:
                raise ValueError(f'"{absolute}" is not a directory')
        else:
            raise ValueError(f'directory "{dirname}" does not exist')

    @property
    def test_files(self) -> str:
        return os.path.expanduser(self.get(CORE, "TEST_FILES"))

    @property
    def test_files_path(self) -> Path:
        return Path(self.test_files)

    @property
    def load_proxy_graphics(self) -> bool:
        return self.get_bool(CORE, "LOAD_PROXY_GRAPHICS", default=True)

    @load_proxy_graphics.setter
    def load_proxy_graphics(self, value: bool) -> None:
        self.set(CORE, "LOAD_PROXY_GRAPHICS", boolstr(value))

    @property
    def store_proxy_graphics(self) -> bool:
        return self.get_bool(CORE, "STORE_PROXY_GRAPHICS", default=True)

    @store_proxy_graphics.setter
    def store_proxy_graphics(self, value: bool) -> None:
        self.set(CORE, "STORE_PROXY_GRAPHICS", boolstr(value))

    @property
    def write_fixed_meta_data_for_testing(self) -> bool:
        # Enable this option to always create same meta data for testing
        # scenarios, e.g. to use a diff like tool to compare DXF documents.
        return self.get_bool(
            CORE, "WRITE_FIXED_META_DATA_FOR_TESTING", default=False
        )

    @write_fixed_meta_data_for_testing.setter
    def write_fixed_meta_data_for_testing(self, state: bool) -> None:
        self.set(CORE, "write_fixed_meta_data_for_testing", boolstr(state))

    @property
    def disable_c_ext(self) -> bool:
        """Disable C-extensions if ``True``. """
        return self.get_bool(CORE, "DISABLE_C_EXT", default=False)

    @property
    def use_matplotlib(self) -> bool:
        """Activate/deactivate Matplotlib support e.g. for testing"""
        return self._use_matplotlib

    @use_matplotlib.setter
    def use_matplotlib(self, state: bool) -> None:
        if MATPLOTLIB:
            self._use_matplotlib = state
        else:  # Matplotlib is not installed
            self._use_matplotlib = False

    @property
    def use_c_ext(self) -> bool:
        """Returns ``True`` if the C-extensions are in use."""
        return self._use_c_ext

    def preserve_proxy_graphics(self, state: bool = True) -> None:
        """Enable/disable proxy graphic load/store support."""
        value = boolstr(state)
        self.set(CORE, "LOAD_PROXY_GRAPHICS", value)
        self.set(CORE, "STORE_PROXY_GRAPHICS", value)

    def print(self):
        """Print current configuration to `stdout`."""
        self._config.write(sys.stdout)

    def write_home_config(self):
        """Write current configuration into file "~/.config/ezdxf/ezdxf.ini" or
        "XDG_CONFIG_HOME/ezdxf/ezdxf.ini".
        """

        home_path = config_home_path()
        if not home_path.exists():
            try:
                home_path.mkdir(parents=True)
            except IOError as e:
                print(str(e))
                return
        try:
            with open(home_path / EZDXF_INI, "wt", encoding="utf8") as fp:
                self.write(fp)
        except IOError as e:
            print(str(e))
        else:
            print(f"created config file: '{fp.name}'")

    def reset(self):
        self._loaded_paths = []
        self._config = default_config()
        self.update_cached_options()

    @staticmethod
    def delete_default_config_files():
        for file in default_config_files():
            if file.exists():
                try:
                    file.unlink()
                    print(f"deleted config file: '{file}'")
                except IOError as e:
                    print(str(e))

    @staticmethod
    def xdg_path(xdg_var: str, directory: str) -> Path:
        return xdg_path(xdg_var, directory)


# Global Options
options = Options()



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXF2000
from ezdxf.math import Vec3, Matrix44, NULLVEC, Z_AXIS
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace

__all__ = ["Ray", "XLine"]

acdb_xline = DefSubclass(
    "AcDbXline",
    {
        "start": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        "unit_vector": DXFAttr(
            11,
            xtype=XType.point3d,
            default=Z_AXIS,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_xline_group_codes = group_code_mapping(acdb_xline)


@register_entity
class XLine(DXFGraphic):
    """DXF XLINE entity"""

    DXFTYPE = "XLINE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_xline)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000
    XLINE_SUBCLASS = "AcDbXline"

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_xline_group_codes, subclass=2, recover=True
            )
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, self.XLINE_SUBCLASS)
        self.dxf.export_dxf_attribs(tagwriter, ["start", "unit_vector"])

    def transform(self, m: Matrix44) -> "XLine":
        """Transform the XLINE/RAY entity by transformation matrix `m` inplace."""
        self.dxf.start = m.transform(self.dxf.start)
        self.dxf.unit_vector = m.transform_direction(
            self.dxf.unit_vector
        ).normalize()
        self.post_transform(m)
        return self

    def translate(self, dx: float, dy: float, dz: float) -> "XLine":
        """Optimized XLINE/RAY translation about `dx` in x-axis, `dy` in
        y-axis and `dz` in z-axis.

        """
        self.dxf.start = Vec3(dx, dy, dz) + self.dxf.start
        # Avoid Matrix44 instantiation if not required:
        if self.is_post_transform_required:
            self.post_transform(Matrix44.translate(dx, dy, dz))
        return self


@register_entity
class Ray(XLine):
    """DXF Ray entity"""

    DXFTYPE = "RAY"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_xline)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000
    XLINE_SUBCLASS = "AcDbRay"



# Copyright (c) 2011-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, List, Sequence
import logging
from ezdxf.lldxf.const import DXFStructureError, DXF12
from .table import (
    Table,
    ViewportTable,
    TextstyleTable,
    LayerTable,
    LinetypeTable,
    AppIDTable,
    ViewTable,
    BlockRecordTable,
    DimStyleTable,
    UCSTable,
)

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        TagWriter,
        Drawing,
        DXFEntity,
        DXFTagStorage,
    )

logger = logging.getLogger("ezdxf")

TABLENAMES = {
    "LAYER": "layers",
    "LTYPE": "linetypes",
    "APPID": "appids",
    "DIMSTYLE": "dimstyles",
    "STYLE": "styles",
    "UCS": "ucs",
    "VIEW": "views",
    "VPORT": "viewports",
    "BLOCK_RECORD": "block_records",
}


class TablesSection:
    def __init__(self, doc: "Drawing", entities: List["DXFEntity"] = None):
        assert doc is not None
        self.doc = doc
        # not loaded tables: table.doc is None
        self.layers = LayerTable()
        self.linetypes = LinetypeTable()
        self.appids = AppIDTable()
        self.dimstyles = DimStyleTable()
        self.styles = TextstyleTable()
        self.ucs = UCSTable()
        self.views = ViewTable()
        self.viewports = ViewportTable()
        self.block_records = BlockRecordTable()

        if entities is not None:
            self._load(entities)
        self._reset_not_loaded_tables()

    def tables(self) -> Sequence[Table]:
        return (
            self.layers,
            self.linetypes,
            self.appids,
            self.dimstyles,
            self.styles,
            self.ucs,
            self.views,
            self.viewports,
            self.block_records,
        )

    def _load(self, entities: List["DXFEntity"]) -> None:
        section_head: "DXFTagStorage" = entities[0]  # type: ignore
        if section_head.dxftype() != "SECTION" or section_head.base_class[
            1
        ] != (2, "TABLES"):
            raise DXFStructureError(
                "Critical structure error in TABLES section."
            )
        del entities[0]  # delete first entity (0, SECTION)

        table_records: List["DXFEntity"] = []
        table_name = None
        for entity in entities:
            if entity.dxftype() == "TABLE":
                if len(table_records):
                    # TABLE entity without preceding ENDTAB entity, should we care?
                    logger.debug(
                        f'Ignore missing ENDTAB entity in table "{table_name}".'
                    )
                    self._load_table(table_name, table_records)  # type: ignore
                table_name = entity.dxf.name
                table_records = [entity]  # collect table head
            elif entity.dxftype() == "ENDTAB":  # do not collect (0, 'ENDTAB')
                self._load_table(table_name, table_records)  # type: ignore
                table_records = (
                    []
                )  # collect entities outside of tables, but ignore it
            else:  # collect table entries
                table_records.append(entity)

        if len(table_records):
            # last ENDTAB entity is missing, should we care?
            logger.debug(
                'Ignore missing ENDTAB entity in table "{}".'.format(table_name)
            )
            self._load_table(table_name, table_records)  # type: ignore

    def _load_table(
        self, name: str, table_entities: Iterable["DXFEntity"]
    ) -> None:
        """
        Load table from tags.

        Args:
            name: table name e.g. VPORT
            table_entities: iterable of table records

        """
        table = getattr(self, TABLENAMES[name])
        if isinstance(table, Table):
            table.load(self.doc, iter(table_entities))

    def _reset_not_loaded_tables(self) -> None:
        for table in self.tables():
            if table.doc is None:
                handle = self.doc.entitydb.next_handle()
                table.reset(self.doc, handle)

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        tagwriter.write_str("  0\nSECTION\n  2\nTABLES\n")
        version = tagwriter.dxfversion
        self.viewports.export_dxf(tagwriter)
        self.linetypes.export_dxf(tagwriter)
        self.layers.export_dxf(tagwriter)
        self.styles.export_dxf(tagwriter)
        self.views.export_dxf(tagwriter)
        self.ucs.export_dxf(tagwriter)
        self.appids.export_dxf(tagwriter)
        self.dimstyles.export_dxf(tagwriter)
        if version > DXF12:
            self.block_records.export_dxf(tagwriter)
        tagwriter.write_tag2(0, "ENDSEC")

    def create_table_handles(self):
        # DXF R12: TABLE does not require a handle and owner tag
        # DXF R2000+: TABLE requires a handle and an owner tag
        for table in self.tables():
            handle = self.doc.entitydb.next_handle()
            table.set_handle(handle)



# Copyright (c) 2018-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, List, Union, Optional, Iterable, Dict, Any
import copy
import logging
from collections import namedtuple

from ezdxf.lldxf import const
from ezdxf.lldxf.types import cast_value
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    group_code_mapping,
)
from ezdxf.lldxf.tags import Tags
from ezdxf.lldxf.const import EXT_MAX_STR_LEN
from ezdxf.math import (
    Vec3,
    NULLVEC,
    X_AXIS,
    Y_AXIS,
    Z_AXIS,
    Matrix44,
    WCSTransform,
    OCSTransform,
    NonUniformScalingError,
)
from ezdxf import colors
from ezdxf.proxygraphic import ProxyGraphicError
from ezdxf.tools.text import safe_string
from ezdxf.tools.handle import safe_handle

from .dxfentity import base_class, SubclassProcessor
from .dxfobj import DXFObject
from .dxfgfx import DXFGraphic, acdb_entity

from .factory import register_entity
from .objectcollection import ObjectCollection

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        TagWriter,
        Drawing,
        DXFNamespace,
        DXFTag,
        DXFEntity,
        BaseLayout,
        EntityQuery,
        Auditor,
    )

__all__ = [
    "MultiLeader",
    "MLeader",
    "MLeaderStyle",
    "MLeaderStyleCollection",
    "MLeaderContext",
    "MTextData",
    "BlockData",
    "LeaderData",
    "LeaderLine",
    "ArrowHeadData",
    "AttribData",
]

logger = logging.getLogger("ezdxf")


# DXF Examples:
# "D:\source\dxftest\CADKitSamples\house design for two family with common staircasedwg.dxf"
# "D:\source\dxftest\CADKitSamples\house design.dxf"

# How to render MLEADER: https://atlight.github.io/formats/dxf-leader.html
# DXF reference:
# http://help.autodesk.com/view/OARX/2018/ENU/?guid=GUID-72D20B8C-0F5E-4993-BEB7-0FCF94F32BE0

acdb_mleader = DefSubclass(
    "AcDbMLeader",
    {
        "version": DXFAttr(270, default=2),
        "style_handle": DXFAttr(340),
        # Theory: Take properties from MLEADERSTYLE,
        # except explicit overridden here:
        "property_override_flags": DXFAttr(90),
        # Bit coded flags from ODA DWG specs pg.158:
        # 1 << 0 = leader_type
        # 1 << 1 = leader_line_color
        # 1 << 2 = leader_linetype_handle
        # 1 << 3 = leader_lineweight
        # 1 << 4 = has_landing
        # 1 << 5 = landing_gap ???
        # 1 << 6 = has_dogleg
        # 1 << 7 = dogleg_length
        # 1 << 8 = arrow_head_handle
        # 1 << 9 = arrow_head_size
        # 1 << 10 = content_type
        # 1 << 11 = text_style_handle
        # 1 << 12 = text_left_attachment_type (of MTEXT)
        # 1 << 13 = text_angle_type (of MTEXT)
        # 1 << 14 = text_alignment_type (of MTEXT)
        # 1 << 15 = text_color (of MTEXT)
        # 1 << 16 = ??? Text height (of MTEXT) ???
        # 1 << 17 = has_text_frame
        # 1 << 18 = ??? Enable use of default MTEXT (from MLEADERSTYLE)
        # 1 << 19 = block_record_handle
        # 1 << 20 = block_color
        # 1 << 21 = block_scale_vector
        # 1 << 22 = block_rotation
        # 1 << 23 = block_connection_type
        # 1 << 24 = ??? Scale ???
        # 1 << 25 = text_right_attachment_type (of MTEXT)
        # 1 << 26 = ??? Text switch alignment type (of MTEXT) ???
        # 1 << 27 = text_attachment_direction (of MTEXT)
        # 1 << 28 = text_top_attachment_type (of MTEXT)
        # 1 << 29 = Text_bottom_attachment_type (of MTEXT)
        # leader_type:
        # 0 = invisible
        # 1 = straight line leader
        # 2 = spline leader
        "leader_type": DXFAttr(170, default=1),
        "leader_line_color": DXFAttr(91, default=colors.BY_BLOCK_RAW_VALUE),
        "leader_linetype_handle": DXFAttr(341),
        "leader_lineweight": DXFAttr(171, default=const.LINEWEIGHT_BYBLOCK),
        "has_landing": DXFAttr(290, default=1),
        "has_dogleg": DXFAttr(291, default=1),
        "dogleg_length": DXFAttr(41, default=8),  # depend on $MEASUREMENT?
        # no handle is default arrow 'closed filled':
        "arrow_head_handle": DXFAttr(342),
        # unscaled arrow head size:
        "arrow_head_size": DXFAttr(42, default=4),  # depend on $MEASUREMENT?
        "content_type": DXFAttr(172, default=2),
        # 0 = None
        # 1 = Block content
        # 2 = MTEXT content
        # 3 = TOLERANCE content
        # Text Content:
        "text_style_handle": DXFAttr(343),
        "text_left_attachment_type": DXFAttr(173, default=1),
        # Values 0-8 are used for the left/right attachment
        # point (attachment direction is horizontal)
        # Attachment point is:
        # 0 = top of top text line
        # 1 = middle of top text line
        # 2 = middle of whole text
        # 3 = middle of bottom text line
        # 4 = bottom of bottom text line
        # 5 = bottom of bottom text line & underline bottom text line
        # 6 = bottom of top text line & underline top text line
        # 7 = bottom of top text line
        # 8 = bottom of top text line & underline all text lines
        "text_right_attachment_type": DXFAttr(95, default=1),  # like 173
        "text_angle_type": DXFAttr(174, default=1),
        # 0 = text angle is equal to last leader line segment angle
        # 1 = text is horizontal
        # 2 = text angle is equal to last leader line segment angle, but potentially
        #     rotated by 180 degrees so the right side is up for readability.
        "text_alignment_type": DXFAttr(175, default=2),
        "text_color": DXFAttr(92, default=colors.BY_BLOCK_RAW_VALUE),
        "has_text_frame": DXFAttr(292, default=0),
        # Block Content:
        "block_record_handle": DXFAttr(344),
        "block_color": DXFAttr(
            93, default=colors.BY_BLOCK_RAW_VALUE
        ),  # raw color
        "block_scale_vector": DXFAttr(
            10, xtype=XType.point3d, default=Vec3(1, 1, 1)
        ),
        "block_rotation": DXFAttr(43, default=0),  # in radians!!!
        "block_connection_type": DXFAttr(176, default=0),
        # 0 = center extents
        # 1 = insertion point
        "is_annotative": DXFAttr(293, default=0),
        # REPEAT "arrow_heads": DXF R2007+
        # arrow_head_index: 94, ???
        # arrow_head_handle: 345
        # END "arrow heads"
        # REPEAT "block attribs" (ATTDEF): DXF R2007+
        # attrib_handle: 330
        # attrib_index: 177, sequential index of the label in the collection
        # attrib_width: 44
        # attrib_text: 302, collision with group code (302, "LEADER{") in context data
        # END "block attribs"
        # Text Content:
        "is_text_direction_negative": DXFAttr(
            294, default=0, dxfversion=const.DXF2007
        ),
        "text_IPE_align": DXFAttr(178, default=0, dxfversion=const.DXF2007),
        "text_attachment_point": DXFAttr(
            179, default=1, dxfversion=const.DXF2007
        ),
        # 1 = left
        # 2 = center
        # 3 = right
        "scale": DXFAttr(45, default=1, dxfversion=const.DXF2007),
        # This defines whether the leaders attach to the left/right of the content
        # block/text, or attach to the top/bottom:
        # 0 = horizontal
        # 1 = vertical
        "text_attachment_direction": DXFAttr(
            271, default=0, dxfversion=const.DXF2010
        ),
        # like 173, but
        # 9 = center
        # 10= underline and center
        "text_bottom_attachment_type": DXFAttr(
            272, default=9, dxfversion=const.DXF2010
        ),
        # like 173, but
        # 9 = center
        # 10= overline and center
        "text_top_attachment_type": DXFAttr(
            273, default=9, dxfversion=const.DXF2010
        ),
        "leader_extend_to_text": DXFAttr(
            295, default=0, dxfversion=const.DXF2013
        ),
    },
)
# The text frame shape is stored in XDATA except for the default rectangle:
# 1001 ACAD
# 1070 <type>
# 2 = rounded rectangle
# 3 = parallelogram
# 4 = triangle
# 5 = square
# 6 = pentagon
# 7 = hexagon
# 8 = octagon
# 9 = circle
# 10= ellipse

acdb_mleader_group_codes = group_code_mapping(acdb_mleader)
CONTEXT_STR = "CONTEXT_DATA{"
LEADER_STR = "LEADER{"
LEADER_LINE_STR = "LEADER_LINE{"
START_CONTEXT_DATA = 300
END_CONTEXT_DATA = 301
START_LEADER = 302
END_LEADER = 303
START_LEADER_LINE = 304
END_LEADER_LINE = 305


def compile_context_tags(
    data: List["DXFTag"], stop_code: int
) -> List[Union["DXFTag", List]]:
    def build_structure(
        tag: "DXFTag", stop: int
    ) -> List[Union["DXFTag", List]]:
        collector = [tag]
        tag = next(tags)
        while tag.code != stop:
            if tag.code == START_LEADER:
                collector.append(build_structure(tag, END_LEADER))  # type: ignore
            # Group code 304 is used also for MTEXT content, therefore always
            # test for group code AND and value string:
            elif tag.code == START_LEADER_LINE and tag.value == LEADER_LINE_STR:
                collector.append(build_structure(tag, END_LEADER_LINE))  # type: ignore
            else:
                collector.append(tag)
            tag = next(tags)
        return collector  # type: ignore

    tags = iter(data)
    return build_structure(next(tags), stop_code)


ArrowHeadData = namedtuple("ArrowHeadData", "index, handle")
AttribData = namedtuple("AttribData", "handle, index, width, text")


@register_entity
class MultiLeader(DXFGraphic):
    DXFTYPE = "MULTILEADER"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_mleader)
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2000

    def __init__(self):
        super().__init__()
        self.context = MLeaderContext()
        self.arrow_heads: List[ArrowHeadData] = []
        self.block_attribs: List[AttribData] = []

    @property
    def has_mtext_content(self) -> bool:
        return self.context.mtext is not None

    def get_mtext_content(self) -> str:
        """Get MTEXT content as string, return "" if MULTILEADER has
        BLOCK content.
        """
        mtext = self.context.mtext
        if mtext is not None:
            return mtext.default_content
        return ""

    def set_mtext_content(self, text: str):
        """Set MTEXT content as string, does nothing if MULTILEADER has
        BLOCK content.
        """
        mtext = self.context.mtext
        if mtext is not None:
            mtext.default_content = safe_string(text, EXT_MAX_STR_LEN)

    @property
    def has_block_content(self) -> bool:
        return self.context.block is not None

    def get_block_content(self) -> Dict[str, str]:
        """Get BLOCK attributes as dictionary of (tag, value) pairs.
        Returns an empty dictionary if MULTILEADER has MTEXT content.
        """
        assert self.doc is not None, "valid DXF document required"
        entitydb = self.doc.entitydb
        tags: Dict[str, str] = dict()
        for attr in self.block_attribs:
            attdef = entitydb.get(attr.handle)
            if attdef is not None:
                tags[attdef.dxf.tag] = attr.text
        return tags

    def set_block_content(self, content: Dict[str, str]):
        """Set BLOCK attributes by a dictionary of (tag, value) pairs.
        Does nothing if MULTILEADER has MTEXT content.
        """
        assert self.doc is not None, "valid DXF document required"
        entitydb = self.doc.entitydb
        tags: Dict[str, str] = dict()
        block_attribs = self.block_attribs
        for index, attr in enumerate(block_attribs):
            attdef = entitydb.get(attr.handle)
            if attdef is not None:
                tag = attdef.dxf.tag
                new_text = safe_string(content.get(tag), EXT_MAX_STR_LEN)
                if new_text is not None:
                    block_attribs[index] = attr._replace(text=new_text)
        return tags

    def _copy_data(self, entity: "DXFEntity") -> None:
        """Copy leaders"""
        assert isinstance(entity, MultiLeader)
        entity.context = copy.deepcopy(self.context)
        entity.arrow_heads = copy.deepcopy(self.arrow_heads)
        entity.block_attribs = copy.deepcopy(self.block_attribs)

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor is None:
            return dxf
        tags = processor.subclass_by_index(2)
        if tags:
            context = self.extract_context_data(tags)
            if context:
                try:
                    self.context = self.load_context(context)
                except const.DXFStructureError:
                    logger.info(
                        f"Context structure error in entity MULTILEADER(#{dxf.handle})"
                    )
        else:
            raise const.DXFStructureError(
                f"missing 'AcDbMLeader' subclass in MULTILEADER(#{dxf.handle})"
            )

        self.arrow_heads = self.extract_arrow_heads(tags)
        self.block_attribs = self.extract_block_attribs(tags)

        processor.fast_load_dxfattribs(
            dxf, acdb_mleader_group_codes, subclass=tags, recover=True
        )
        return dxf

    @staticmethod
    def extract_context_data(tags: Tags) -> List["DXFTag"]:
        start, end = None, None
        context_data = []
        for index, tag in enumerate(tags):
            if tag.code == START_CONTEXT_DATA:
                start = index
            elif tag.code == END_CONTEXT_DATA:
                end = index + 1

        if start and end:
            context_data = tags[start:end]
            # Remove context data!
            del tags[start:end]
        return context_data

    @staticmethod
    def load_context(data: List["DXFTag"]) -> "MLeaderContext":
        try:
            context = compile_context_tags(data, END_CONTEXT_DATA)
        except StopIteration:
            raise const.DXFStructureError
        else:
            return MLeaderContext.load(context)

    @staticmethod
    def extract_arrow_heads(data: Tags) -> List[ArrowHeadData]:
        def store_head():
            heads.append(
                ArrowHeadData(
                    collector.get(94, 0),  # arrow head index
                    collector.get(345, "0"),  # arrow head handle
                )
            )
            collector.clear()

        heads: List[ArrowHeadData] = []
        try:
            start = data.tag_index(94)
        except const.DXFValueError:
            return heads

        end = start
        collector = dict()
        for code, value in data.collect_consecutive_tags({94, 345}, start):
            end += 1
            collector[code] = value
            if code == 345:
                store_head()

        # Remove processed tags:
        del data[start:end]
        return heads

    @staticmethod
    def extract_block_attribs(data: Tags) -> List[AttribData]:
        def store_attrib():
            attribs.append(
                AttribData(
                    collector.get(330, "0"),  # ATTDEF handle
                    collector.get(177, 0),  # ATTDEF index
                    collector.get(44, 1.0),  # ATTDEF width
                    collector.get(302, ""),  # ATTDEF text (content)
                )
            )
            collector.clear()

        attribs: List[AttribData] = []
        try:
            start = data.tag_index(330)
        except const.DXFValueError:
            return attribs

        end = start
        collector: Dict[int, Any] = dict()
        for code, value in data.collect_consecutive_tags(
            {330, 177, 44, 302}, start
        ):
            end += 1
            if code == 330 and len(collector):
                store_attrib()
            collector[code] = value
        if len(collector):
            store_attrib()

        # Remove processed tags:
        del data[start:end]
        return attribs

    def preprocess_export(self, tagwriter: "TagWriter") -> bool:
        if self.context.is_valid:
            return True
        else:
            logger.debug(
                f"Ignore {str(self)} at DXF export, invalid context data."
            )
            return False

    def export_entity(self, tagwriter: "TagWriter") -> None:
        def write_handle_if_exist(code: int, name: str):
            handle = dxf.get(name)
            if handle is not None:
                write_tag2(code, handle)

        super().export_entity(tagwriter)
        dxf = self.dxf
        version = tagwriter.dxfversion
        write_tag2 = tagwriter.write_tag2

        write_tag2(100, acdb_mleader.name)
        write_tag2(270, dxf.version)
        self.context.export_dxf(tagwriter)

        # Export common MLEADER tags:
        # Don't use dxf.export_dxf_attribs() - all attributes should be written
        # even if equal to the default value:
        write_tag2(340, dxf.style_handle)
        write_tag2(90, dxf.property_override_flags)
        write_tag2(170, dxf.leader_type)
        write_tag2(91, dxf.leader_line_color)
        write_tag2(341, dxf.leader_linetype_handle)
        write_tag2(171, dxf.leader_lineweight)
        write_tag2(290, dxf.has_landing)
        write_tag2(291, dxf.has_dogleg)
        write_tag2(41, dxf.dogleg_length)
        # arrow_head_handle is None for default arrow 'closed filled':
        write_handle_if_exist(342, "arrow_head_handle")
        write_tag2(42, dxf.arrow_head_size)
        write_tag2(172, dxf.content_type)
        write_tag2(343, dxf.text_style_handle)  # mandatory!
        write_tag2(173, dxf.text_left_attachment_type)
        write_tag2(95, dxf.text_right_attachment_type)
        write_tag2(174, dxf.text_angle_type)
        write_tag2(175, dxf.text_alignment_type)
        write_tag2(92, dxf.text_color)
        write_tag2(292, dxf.has_text_frame)

        write_handle_if_exist(344, "block_record_handle")
        write_tag2(93, dxf.block_color)
        tagwriter.write_vertex(10, dxf.block_scale_vector)
        write_tag2(43, dxf.block_rotation)
        write_tag2(176, dxf.block_connection_type)
        write_tag2(293, dxf.is_annotative)
        if version >= const.DXF2007:
            self.export_arrow_heads(tagwriter)
            self.export_block_attribs(tagwriter)
            write_tag2(294, dxf.is_text_direction_negative)
            write_tag2(178, dxf.text_IPE_align)
            write_tag2(179, dxf.text_attachment_point)
            write_tag2(45, dxf.scale)

        if version >= const.DXF2010:
            write_tag2(271, dxf.text_attachment_direction)
            write_tag2(272, dxf.text_bottom_attachment_type)
            write_tag2(273, dxf.text_top_attachment_type)

        if version >= const.DXF2013:
            write_tag2(295, dxf.leader_extend_to_text)

    def export_arrow_heads(self, tagwriter: "TagWriter") -> None:
        for index, handle in self.arrow_heads:
            tagwriter.write_tag2(94, index)
            tagwriter.write_tag2(345, handle)

    def export_block_attribs(self, tagwriter: "TagWriter") -> None:
        for attrib in self.block_attribs:
            tagwriter.write_tag2(330, attrib.handle)
            tagwriter.write_tag2(177, attrib.index)
            tagwriter.write_tag2(44, attrib.width)
            tagwriter.write_tag2(302, safe_string(attrib.text, EXT_MAX_STR_LEN))

    def virtual_entities(self) -> Iterable["DXFGraphic"]:
        """Yields the graphical representation of MULTILEADER as virtual DXF
        primitives.

        This entities are located at the original location, but are not stored
        in the entity database, have no handle and are not assigned to any
        layout.

        """
        return self.__virtual_entities__()

    def explode(self, target_layout: "BaseLayout" = None) -> "EntityQuery":
        """Explode MULTILEADER as DXF primitives into target layout,
        if target layout is ``None``, the target layout is the layout of the
        source entity.

        Returns an :class:`~ezdxf.query.EntityQuery` container with all DXF
        primitives.

        Args:
            target_layout: target layout for the DXF primitives, ``None`` for
                same layout as the source entity.

        """
        from ezdxf.explode import explode_entity

        return explode_entity(self, target_layout)

    def __virtual_entities__(self) -> Iterable["DXFGraphic"]:
        """Support for "VirtualEntities" protocol."""
        from ezdxf.render import mleader

        try:
            return mleader.virtual_entities(self, proxy_graphic=True)
        except ProxyGraphicError:
            return mleader.virtual_entities(self, proxy_graphic=False)

    def __referenced_blocks__(self) -> Iterable[str]:
        """Support for "ReferencedBlocks" protocol."""
        # 1. overridden MLEADERSTYLE attributes
        for name in ("block_record_handle", "arrow_head_handle"):
            handle = self.dxf.get(name, None)
            if handle is not None:
                yield handle

        # 2. arrows heads
        for arrow_head in self.arrow_heads:
            yield arrow_head.handle

        # 3. MLeader context
        if self.context.block is not None:
            handle = self.context.block.block_record_handle
            if handle is not None:
                yield handle

    def transform(self, m: "Matrix44") -> "MultiLeader":
        """Transform the MULTILEADER entity by transformation matrix `m` inplace.

        Non uniform scaling is not supported.

        Args:
            m: transformation :class:`~ezdxf.math.Matrix44`

        Raises:
            NonUniformScalingError: for non uniform scaling

        """
        dxf = self.dxf
        context = self.context

        wcs = WCSTransform(m)
        if not wcs.has_uniform_xy_scaling:
            # caller has to catch this exception and explode the MULTILEADER
            raise NonUniformScalingError(
                "MULTILEADER does not support non uniform scaling"
            )
        if (
            abs(context.plane_x_axis.z) > 1e-12
            or abs(context.plane_y_axis.z) > 1e-12
        ):
            # check only if not parallel to xy-plane
            if not wcs.has_uniform_xyz_scaling:
                # caller has to catch this exception and explode the MULTILEADER
                raise NonUniformScalingError(
                    "MULTILEADER does not support non uniform scaling"
                )
        context.transform(wcs)
        # copy redundant attributes from sub-structures:
        dxf.arrow_head_size = context.arrow_head_size
        dxf.scale = context.scale
        dxf.dogleg_length *= wcs.uniform_scale
        if context.block:
            dxf.block_rotation = context.block.rotation
            dxf.block_scale_vector = context.block.scale

        # ArrowHeadData: no transformation needed
        # AttribData: no transformation needed
        self.post_transform(m)

        # The proxy graphic is stored in absolute WCS coordinates and does not
        # represent the new transformed geometry.
        self.update_proxy_graphic()
        return self

    def update_proxy_graphic(self):
        # Maybe this is supported in the future - its very unlikeLy!
        self.proxy_graphic = None


@register_entity
class MLeader(MultiLeader):  # same entity different name
    DXFTYPE = "MLEADER"


class MLeaderContext:
    ATTRIBS = {
        40: "scale",
        10: "base_point",
        41: "char_height",
        140: "arrow_head_size",
        145: "landing_gap_size",
        174: "left_attachment",
        175: "right_attachment",
        176: "text_align_type",
        177: "attachment_type",
        110: "plane_origin",
        111: "plane_x_axis",
        112: "plane_y_axis",
        297: "plane_normal_reversed",
        272: "top_attachment",
        273: "bottom_attachment",
    }

    def __init__(self):
        self.leaders: List["LeaderData"] = []
        self.scale: float = 1.0  # overall scale

        # MTEXT base point: is not the MTEXT insertion point!
        # HORIZONTAL leader attachment:
        # the "base_point" is always the start point of the leader on the LEFT side
        # of the MTEXT, regardless of alignment and which side leaders are attached.
        # VERTICAL leader attachment:
        # the "base_point" is always the start point of the leader on the BOTTOM
        # side of the MTEXT, regardless of alignment and which side leaders are
        # attached.

        # BLOCK base point: is not the BLOCK insertion point!
        # HORIZONTAL leader attachment:
        # Strange results, setting the "base_point" to the left center of the BLOCK,
        # regardless which side leaders are attached, seems to be reasonable.
        # VERTICAL leader attachment: not supported by BricsCAD
        self.base_point: Vec3 = NULLVEC
        self.char_height = 4.0  # scaled char height!
        self.arrow_head_size = 4.0
        self.landing_gap_size = 2.0
        self.left_attachment = 1
        self.right_attachment = 1
        self.text_align_type = 0  # 0=left, 1=center, 2=right
        self.attachment_type = 0  # 0=content extents, 1=insertion point
        self.mtext: Optional[MTextData] = None
        self.block: Optional[BlockData] = None
        self.plane_origin: Vec3 = NULLVEC
        self.plane_x_axis: Vec3 = X_AXIS
        self.plane_y_axis: Vec3 = Y_AXIS
        self.plane_normal_reversed: int = 0
        self.top_attachment = 9
        self.bottom_attachment = 9

    @classmethod
    def load(cls, context: List[Union["DXFTag", List]]) -> "MLeaderContext":
        assert context[0] == (START_CONTEXT_DATA, CONTEXT_STR)
        ctx = cls()
        content = None
        for tag in context:
            if isinstance(tag, list):  # Leader()
                ctx.leaders.append(LeaderData.load(tag))
                continue
            # parse context tags
            code, value = tag
            if content:
                if content.parse(code, value):
                    continue
                else:
                    content = None

            if code == 290 and value == 1:
                content = MTextData()
                ctx.mtext = content
            elif code == 296 and value == 1:
                content = BlockData()  # type: ignore
                ctx.block = content
            else:
                name = MLeaderContext.ATTRIBS.get(code)
                if name:
                    ctx.__setattr__(name, cast_value(code, value))
        return ctx

    @property
    def is_valid(self) -> bool:
        return True

    @property
    def plane_z_axis(self) -> Vec3:
        z_axis = self.plane_x_axis.cross(self.plane_y_axis).normalize()
        if self.plane_normal_reversed:
            z_axis = -z_axis
        return z_axis

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex
        write_tag2(START_CONTEXT_DATA, CONTEXT_STR)
        # All MLeaderContext tags:
        write_tag2(40, self.scale)
        write_vertex(10, self.base_point)
        write_tag2(41, self.char_height)
        write_tag2(140, self.arrow_head_size)
        write_tag2(145, self.landing_gap_size)
        write_tag2(174, self.left_attachment)
        write_tag2(175, self.right_attachment)
        write_tag2(176, self.text_align_type)
        write_tag2(177, self.attachment_type)

        if self.mtext:
            write_tag2(290, 1)  # has mtext content
            self.mtext.export_dxf(tagwriter)
        else:
            write_tag2(290, 0)

        if self.block:
            write_tag2(296, 1)  # has block content
            self.block.export_dxf(tagwriter)
        else:
            write_tag2(296, 0)

        write_vertex(110, self.plane_origin)
        write_vertex(111, self.plane_x_axis)
        write_vertex(112, self.plane_y_axis)
        write_tag2(297, self.plane_normal_reversed)

        # Export Leader and LiederLine objects:
        for leader in self.leaders:
            leader.export_dxf(tagwriter)

        # Additional MLeaderContext tags:
        if tagwriter.dxfversion >= const.DXF2010:
            write_tag2(272, self.top_attachment)
            write_tag2(273, self.bottom_attachment)
        write_tag2(END_CONTEXT_DATA, "}")

    def transform(self, wcs: WCSTransform) -> None:
        m = wcs.m
        scale = wcs.uniform_scale
        self.scale *= scale
        self.base_point = m.transform(self.base_point)
        self.char_height *= scale
        self.arrow_head_size *= scale
        self.landing_gap_size *= scale
        self.plane_origin = m.transform(
            self.plane_origin
        )  # confirmed by BricsCAD
        self.plane_x_axis = m.transform_direction(
            self.plane_x_axis, normalize=True
        )
        self.plane_y_axis = m.transform_direction(
            self.plane_y_axis, normalize=True
        )
        self.plane_normal_reversed = 0
        z_axis = m.transform_direction(Z_AXIS, normalize=True)
        if z_axis.isclose(-self.plane_z_axis):  # reversed z-axis?
            self.plane_normal_reversed = 1

        for leader in self.leaders:
            leader.transform(wcs)
        if self.mtext is not None:
            self.mtext.transform(wcs)
        if self.block is not None:
            self.block.transform(wcs)

    def set_scale(self, value: float):
        try:
            conversion_factor = value / self.scale
        except ZeroDivisionError:
            return
        self.scale = value
        self.char_height *= conversion_factor
        self.arrow_head_size *= conversion_factor
        self.landing_gap_size *= conversion_factor


class MTextData:
    ATTRIBS = {
        304: "default_content",
        11: "extrusion",
        340: "style_handle",
        12: "insert",
        13: "text_direction",
        42: "rotation",
        43: "width",
        44: "defined_height",
        45: "line_spacing_factor",
        170: "line_spacing_style",
        90: "color",
        171: "alignment",
        172: "flow_direction",
        91: "bg_color",
        141: "bg_scale_factor",
        92: "bg_transparency",
        291: "use_window_bg_color",
        292: "has_bg_fill",
        173: "column_type",
        293: "use_auto_height",
        142: "column_width",
        143: "column_gutter_width",
        294: "column_flow_reversed",
        144: "column_sizes",  # multiple values
        295: "use_word_break",
    }

    def __init__(self):
        self.default_content: str = ""
        self.extrusion: Vec3 = Z_AXIS
        self.style_handle: str = "0"  # handle of TextStyle() table entry
        self.insert: Vec3 = NULLVEC
        self.text_direction: Vec3 = X_AXIS  # text direction
        self.rotation: float = 0.0  # in radians!
        self.width: float = 0.0  # MTEXT width, not scaled
        self.defined_height: float = 0.0  # defined column height, not scaled
        self.line_spacing_factor: float = 1.0
        self.line_spacing_style: int = 1  # 1=at least, 2=exactly
        self.color: int = colors.BY_BLOCK_RAW_VALUE
        self.alignment: int = 1  # 1=left, 2=center, 3=right
        self.flow_direction: int = 1  # 1=horiz, 3=vert, 6=by style
        self.bg_color: int = colors.WINDOW_BG_RAW_VALUE
        self.bg_scale_factor: float = 1.5
        self.bg_transparency: int = 0
        self.use_window_bg_color: int = 0
        self.has_bg_fill: int = 0
        self.column_type: int = 0  # unknown values
        self.use_auto_height: int = 0
        self.column_width: float = 0.0  # not scaled
        self.column_gutter_width: float = 0.0  # not scaled
        self.column_flow_reversed: int = 0
        self.column_sizes: List[float] = []  # heights?, not scaled
        self.use_word_break: int = 1

    def parse(self, code: int, value) -> bool:
        # return True if data belongs to mtext else False (end of mtext section)
        if code == 144:
            self.column_sizes.append(float(value))
            return True
        attrib = MTextData.ATTRIBS.get(code)
        if attrib:
            self.__setattr__(attrib, cast_value(code, value))
        return bool(attrib)

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex
        write_tag2(304, safe_string(self.default_content, EXT_MAX_STR_LEN))
        write_vertex(11, self.extrusion)
        write_tag2(340, safe_handle(self.style_handle))
        write_vertex(12, self.insert)
        write_vertex(13, self.text_direction)
        write_tag2(42, self.rotation)
        write_tag2(43, self.width)
        write_tag2(44, self.defined_height)
        write_tag2(45, self.line_spacing_factor)
        write_tag2(170, self.line_spacing_style)
        write_tag2(90, self.color)
        write_tag2(171, self.alignment)
        write_tag2(172, self.flow_direction)
        write_tag2(91, self.bg_color)
        write_tag2(141, self.bg_scale_factor)
        write_tag2(92, self.bg_transparency)
        write_tag2(291, self.use_window_bg_color)
        write_tag2(292, self.has_bg_fill)
        write_tag2(173, self.column_type)
        write_tag2(293, self.use_auto_height)
        write_tag2(142, self.column_width)
        write_tag2(143, self.column_gutter_width)
        write_tag2(294, self.column_flow_reversed)
        for size in self.column_sizes:
            write_tag2(144, size)
        write_tag2(295, self.use_word_break)

    def transform(self, wcs: WCSTransform) -> None:
        # MTEXT is not really an OCS entity!
        m = wcs.m
        ocs = OCSTransform(self.extrusion, m)  # source extrusion!
        self.extrusion = ocs.new_extrusion  # process like an OCS entity!
        self.insert = m.transform(self.insert)  # WCS
        self.text_direction = wcs.m.transform_direction(  # WCS
            self.text_direction, normalize=True
        )
        # don't use rotation ;)
        self.rotation = ocs.transform_angle(self.rotation)
        scale = wcs.uniform_scale
        self.width *= scale
        self.defined_height *= scale
        self.column_width *= scale
        self.column_gutter_width *= scale
        self.column_sizes = [size * scale for size in self.column_sizes]

    def apply_conversion_factor(self, conversion_factor: float):
        # conversion_factor: convert from an old scaling to a new scaling
        self.width *= conversion_factor
        self.defined_height *= conversion_factor
        self.column_width *= conversion_factor
        self.column_gutter_width *= conversion_factor
        self.column_sizes = [h * conversion_factor for h in self.column_sizes]


class BlockData:
    ATTRIBS = {
        341: "block_record_handle",
        14: "extrusion",
        15: "insert",
        16: "scale",
        46: "rotation",
        93: "color",
    }

    def __init__(self):
        self.block_record_handle = None
        self.extrusion: Vec3 = Z_AXIS
        self.insert: Vec3 = NULLVEC
        self.scale: Vec3 = Vec3(1, 1, 1)
        self.rotation: float = 0  # in radians!
        self.color: int = colors.BY_BLOCK_RAW_VALUE
        # The transformation matrix is stored in transposed order
        # of ezdxf.math.Matrix44()!
        self._matrix: List[float] = []  # group code 47 x 16

    @property
    def matrix44(self) -> Matrix44:
        if len(self._matrix) == 16:
            m = Matrix44(self._matrix)
        else:
            m = Matrix44()
        m.transpose()
        return m

    @matrix44.setter
    def matrix44(self, m: Matrix44) -> None:
        m = m.copy()
        m.transpose()
        self._matrix = list(m)

    def parse(self, code: int, value) -> bool:
        attrib = BlockData.ATTRIBS.get(code)
        if attrib:
            self.__setattr__(attrib, cast_value(code, value))
        elif code == 47:
            self._matrix.append(float(value))
        else:
            return False
        # return True if data belongs to block else False (end of block section)
        return True

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex
        if self.block_record_handle:
            write_tag2(341, self.block_record_handle)
        else:
            # Do not write None, but "0" is also not valid!
            # DXF structure error should be detected before export.
            write_tag2(341, "0")
        write_vertex(14, self.extrusion)
        write_vertex(15, self.insert)
        write_vertex(16, self.scale)
        write_tag2(46, self.rotation)
        write_tag2(93, self.color)
        m = self._matrix
        if len(m) != 16:
            # The content of this matrix is not used for block transformation,
            # not by AutoCAD nor by BricsCAD, but BricsCAD gets irritated
            # if this matrix is missing. AutoCAD is fine without the presence
            # of this matrix.
            m = list(Matrix44())  # identity matrix
        for value in m:
            write_tag2(47, value)

    def transform(self, wcs: WCSTransform) -> None:
        # INSERT is an OCS entity, but the insert point is WCS!!!
        m = wcs.m
        ocs = OCSTransform(self.extrusion, m)  # source extrusion!
        self.extrusion = ocs.new_extrusion
        self.insert = m.transform(self.insert)  # WCS coordinates!!!
        self.scale = ocs.transform_scale_vector(self.scale)
        self.rotation = ocs.transform_angle(self.rotation)
        self.matrix44 = self.matrix44 * m

    def apply_conversion_factor(self, conversion_factor: float):
        # conversion_factor: convert from an old scaling to a new scaling
        self.scale *= conversion_factor


class LeaderData:
    def __init__(self):
        self.lines: List["LeaderLine"] = []
        # has_last_leader_line:
        # in AutoCAD the leader is invisible if set to 0
        # BricsCAD ignores this flag
        self.has_last_leader_line: int = 0  # group code 290,
        self.has_dogleg_vector: int = 0  # group code 291
        self.last_leader_point: Vec3 = NULLVEC  # group code (10, 20, 30) in WCS
        self.dogleg_vector: Vec3 = X_AXIS  # group code (11, 21, 31) in WCS
        self.dogleg_length: float = 1.0  # group code 40
        self.index: int = 0  # group code 90

        # 0=horizontal; 1=vertical
        self.attachment_direction: int = 0  # group code 271, R2010+
        self.breaks: List[
            Vec3
        ] = []  # group code 12, 13 - multiple breaks possible!

    @property
    def has_horizontal_attachment(self) -> bool:
        return not bool(self.attachment_direction)

    @classmethod
    def load(cls, context: List[Union["DXFTag", List]]):
        assert context[0] == (START_LEADER, LEADER_STR)
        leader = cls()
        for tag in context:
            if isinstance(tag, list):  # LeaderLine()
                leader.lines.append(LeaderLine.load(tag))
                continue

            code, value = tag
            if code == 290:
                leader.has_last_leader_line = value
            elif code == 291:
                leader.has_dogleg_vector = value
            elif code == 10:
                leader.last_leader_point = Vec3(value)
            elif code == 11:
                leader.dogleg_vector = Vec3(value)
            elif code == 40:
                leader.dogleg_length = value
            elif code == 90:
                leader.index = value
            elif code == 271:
                leader.attachment_direction = value
            elif code in (12, 13):
                leader.breaks.append(Vec3(value))

        return leader

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex

        write_tag2(START_LEADER, LEADER_STR)
        write_tag2(290, self.has_last_leader_line)
        write_tag2(291, self.has_dogleg_vector)
        write_vertex(10, self.last_leader_point)
        write_vertex(11, self.dogleg_vector)

        code = 0
        for vertex in self.breaks:
            # write alternate group code 12 and 13
            write_vertex(12 + code, vertex)
            code = 1 - code
        write_tag2(90, self.index)
        write_tag2(40, self.dogleg_length)

        # Export leader lines:
        for line in self.lines:
            line.export_dxf(tagwriter)

        if tagwriter.dxfversion >= const.DXF2010:
            write_tag2(271, self.attachment_direction)
        write_tag2(END_LEADER, "}")

    def transform(self, wcs: WCSTransform) -> None:
        m = wcs.m
        self.last_leader_point = m.transform(self.last_leader_point)
        try:
            dog_leg = m.transform_direction(
                self.dogleg_vector.normalize(self.dogleg_length)
            )
        except ZeroDivisionError:  # dogleg_vector is NULL
            dog_leg = m.transform_direction(Vec3(self.dogleg_length, 0, 0))
        self.dogleg_vector = dog_leg.normalize()
        self.dogleg_length = dog_leg.magnitude
        self.breaks = list(m.transform_vertices(self.breaks))
        for leader_line in self.lines:
            leader_line.transform(wcs)


class LeaderLine:
    def __init__(self):
        self.vertices: List[Vec3] = []  # WCS coordinates
        self.breaks: List[Union[int, Vec3]] = []
        # Breaks: 90, 11, 12, [11, 12, ...] [, 90, 11, 12 [11, 12, ...]]
        # group code 90 = break index
        # group code 11 = start vertex of break
        # group code 12 = end vertex of break
        # multiple breaks per index possible
        self.index: int = 0  # group code 91
        self.color: int = colors.BY_BLOCK_RAW_VALUE  # group code 92
        # R2010+: override properties see ODA DWG pg. 214-215

    @classmethod
    def load(cls, tags: List["DXFTag"]):
        assert tags[0] == (START_LEADER_LINE, LEADER_LINE_STR)
        line = LeaderLine()
        vertices = line.vertices
        breaks = []
        for code, value in tags:
            if code == 10:
                vertices.append(Vec3(value))
            elif code in (90, 11, 12):
                breaks.append(cast_value(code, value))
            elif code == 91:
                line.index = value
            elif code == 92:
                line.color = value
        if breaks:
            line.breaks = breaks
        return line

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex

        write_tag2(START_LEADER_LINE, LEADER_LINE_STR)
        for vertex in self.vertices:
            write_vertex(10, vertex)
        if self.breaks:
            code = 0
            for value in self.breaks:
                if isinstance(value, int):
                    # break index
                    write_tag2(90, value)
                else:
                    # 11 .. start vertex of break
                    # 12 .. end vertex of break
                    write_vertex(11 + code, value)
                    code = 1 - code
        write_tag2(91, self.index)
        write_tag2(92, self.color)
        write_tag2(END_LEADER_LINE, "}")

    def transform(self, wcs: WCSTransform) -> None:
        m = wcs.m
        self.vertices = list(m.transform_vertices(self.vertices))
        if self.breaks:
            breaks: List[Union[int, Vec3]] = []
            for value in self.breaks:
                if isinstance(value, Vec3):
                    breaks.append(m.transform(value))
                else:  # break index
                    breaks.append(value)
            self.breaks = breaks


acdb_mleader_style = DefSubclass(
    "AcDbMLeaderStyle",
    {
        "unknown1": DXFAttr(179, default=2),
        "content_type": DXFAttr(170, default=2),
        "draw_mleader_order_type": DXFAttr(171, default=1),
        "draw_leader_order_type": DXFAttr(172, default=0),
        "max_leader_segments_points": DXFAttr(90, default=2),
        "first_segment_angle_constraint": DXFAttr(40, default=0.0),
        "second_segment_angle_constraint": DXFAttr(41, default=0.0),
        # leader_type:
        # 0 = invisible
        # 1 = straight line leader
        # 2 = spline leader
        "leader_type": DXFAttr(173, default=1),
        "leader_line_color": DXFAttr(91, default=colors.BY_BLOCK_RAW_VALUE),
        "leader_linetype_handle": DXFAttr(340),
        "leader_lineweight": DXFAttr(92, default=-2),
        "has_landing": DXFAttr(290, default=1),
        "landing_gap": DXFAttr(42, default=2.0),
        "has_dogleg": DXFAttr(291, default=1),
        "dogleg_length": DXFAttr(43, default=8),
        "name": DXFAttr(3, default="Standard"),  # description text
        # no handle is default arrow 'closed filled':
        "arrow_head_handle": DXFAttr(341),
        "arrow_head_size": DXFAttr(44, default=4),
        "default_text_content": DXFAttr(300, default=""),
        "text_style_handle": DXFAttr(342),
        "text_left_attachment_type": DXFAttr(174, default=1),
        "text_angle_type": DXFAttr(175, default=1),
        "text_alignment_type": DXFAttr(176, default=0),
        "text_right_attachment_type": DXFAttr(178, default=1),
        "text_color": DXFAttr(93, default=colors.BY_BLOCK_RAW_VALUE),
        "char_height": DXFAttr(45, default=4),
        "has_text_frame": DXFAttr(292, default=0),
        "text_align_always_left": DXFAttr(297, default=0),
        "align_space": DXFAttr(46, default=4),
        "has_block_scaling": DXFAttr(293),
        "block_record_handle": DXFAttr(343),
        "block_color": DXFAttr(94, default=colors.BY_BLOCK_RAW_VALUE),
        "block_scale_x": DXFAttr(47, default=1),
        "block_scale_y": DXFAttr(49, default=1),
        "block_scale_z": DXFAttr(140, default=1),
        "has_block_rotation": DXFAttr(294, default=1),
        "block_rotation": DXFAttr(141, default=0),
        "block_connection_type": DXFAttr(177, default=0),
        "scale": DXFAttr(142, default=1),
        "overwrite_property_value": DXFAttr(295, default=0),
        "is_annotative": DXFAttr(296, default=0),
        "break_gap_size": DXFAttr(143, default=3.75),
        # 0 = Horizontal; 1 = Vertical:
        "text_attachment_direction": DXFAttr(271, default=0),
        # 9 = Center; 10 = Underline and Center:
        "text_bottom_attachment_type": DXFAttr(272, default=9),
        # 9 = Center; 10 = Overline and Center:
        "text_top_attachment_type": DXFAttr(273, default=9),
        "unknown2": DXFAttr(298, optional=True),  # boolean flag ?
    },
)
acdb_mleader_style_group_codes = group_code_mapping(acdb_mleader_style)


@register_entity
class MLeaderStyle(DXFObject):
    DXFTYPE = "MLEADERSTYLE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_mleader_style)
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2000

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_mleader_style_group_codes, subclass=1
            )
        return dxf

    def set_mtext_style(self, name: str) -> None:
        assert self.doc is not None, "valid DXF document required"
        style = self.doc.styles.get(name)
        if style is not None:
            self.dxf.text_style_handle = style.dxf.handle
        else:
            raise ValueError(f"text style '{name}' does not exist")

    def set_leader_properties(
        self,
        color: Union[int, colors.RGB] = colors.BYBLOCK,
        linetype: str = "BYBLOCK",
        lineweight: int = const.LINEWEIGHT_BYBLOCK,
        leader_type: int = 1,
    ):
        assert self.doc is not None, "valid DXF document required"
        self.dxf.leader_line_color = colors.encode_raw_color(color)  # type: ignore
        linetype_ = self.doc.linetypes.get(linetype)
        if linetype_ is None:
            raise ValueError(f"invalid linetype name '{linetype}'")
        self.dxf.leader_linetype_handle = linetype_.dxf.handle
        self.dxf.leader_lineweight = int(lineweight)
        self.dxf.leader_type = int(leader_type)

    def set_arrow_head(self, name: str = ""):
        from ezdxf.render.arrows import ARROWS

        assert self.doc is not None, "valid DXF document required"
        if name:
            self.dxf.arrow_head_handle = ARROWS.arrow_handle(
                self.doc.blocks, name
            )
        else:
            # empty string is the default "closed filled" arrow,
            # no handle needed
            del self.dxf.arrow_head_handle

    def export_entity(self, tagwriter: "TagWriter") -> None:
        super().export_entity(tagwriter)
        tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_mleader_style.name)
        self.dxf.export_dxf_attribs(
            tagwriter, acdb_mleader_style.attribs.keys()
        )

    def __referenced_blocks__(self) -> Iterable[str]:
        """Support for "ReferencedBlocks" protocol."""
        for name in ("block_record_handle", "arrow_head_handle"):
            handle = self.dxf.get(name, None)
            if handle is not None:
                yield handle

    def audit(self, auditor: "Auditor") -> None:
        from ezdxf.audit import AuditError

        if not self.is_alive:
            return
        super().audit(auditor)
        entitydb = auditor.entitydb
        dxf = self.dxf
        name = str(self)
        handle = dxf.get("text_style_handle", None)
        if handle is not None and handle not in entitydb:
            standard = auditor.doc.styles.get("Standard")
            if standard is not None:
                self.dxf.text_style_handle = standard.dxf.handle
                auditor.fixed_error(
                    AuditError.UNDEFINED_TEXT_STYLE,
                    f"{name}: text_style_handle={handle} is not valid, replaced by "
                    f"'Standard' text style",
                    self
                )
            else:
                logger.warning("required text style 'Standard' does not exist")
                self.dxf.discard("text_style_handle")
                auditor.fixed_error(
                    AuditError.UNDEFINED_TEXT_STYLE,
                    f"{name}: removed invalid text_style_handle={handle}",
                    self
                )
        for attrib in ("arrow_head_handle", "block_record_handle"):
            handle = dxf.get(attrib)
            if handle is not None and handle not in entitydb:
                dxf.discard(attrib)
                auditor.fixed_error(
                    AuditError.UNDEFINED_BLOCK,
                    f"{name}: removed invalid {attrib}={handle}",
                    self
                )


class MLeaderStyleCollection(ObjectCollection):
    def __init__(self, doc: "Drawing"):
        super().__init__(
            doc, dict_name="ACAD_MLEADERSTYLE", object_type="MLEADERSTYLE"
        )
        self.create_required_entries()

    def create_required_entries(self) -> None:
        if "Standard" not in self:
            mleader_style = self.new("Standard")
            # set standard text style
            text_style = self.doc.styles.get("Standard")
            mleader_style.dxf.text_style_handle = text_style.dxf.handle



# Copyright (c) 2020-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, Tuple, Dict, Union, List
import heapq

if TYPE_CHECKING:
    from ezdxf.eztypes import DXFGraphic

__all__ = ["ascending", "descending"]

# ODA DWG Specs: 2.13. Handle References
# COUNTER is 4 bits, which allows handles up to 16 * 1 byte = 128-bit
# Example for 128-bit handles: "CADKitSamples\AEC Plan Elev Sample.dxf"
MAX_HANDLE = "FFFFFFFFFFFFFFFF"
NULL_HANDLE = "0"


def ascending(
    entities: Iterable["DXFGraphic"],
    mapping: Union[Dict, Iterable[Tuple[str, str]]] = None,
) -> Iterable["DXFGraphic"]:
    """Yields entities in ascending handle order.

    The sort handle doesn't have to be the entity handle, every entity handle
    in `mapping` will be replaced by the given sort handle, `mapping` is an
    iterable of 2-tuples (entity_handle, sort_handle) or a
    dict (entity_handle, sort_handle). Entities with equal sort handles show
    up in source entities order.

    Args:
        entities: iterable of :class:`DXFGraphic` objects
        mapping: iterable of 2-tuples (entity_handle, sort_handle) or a
            handle mapping as dict.

    """
    mapping = dict(mapping) if mapping else {}
    heap = _build(entities, mapping, +1)
    return _sorted(heap)


def descending(
    entities: Iterable["DXFGraphic"],
    mapping: Union[Dict, Iterable[Tuple[str, str]]] = None,
) -> Iterable["DXFGraphic"]:
    """Yields entities in descending handle order.

    The sort handle doesn't have to be the entity handle, every entity handle
    in `mapping` will be replaced by the given sort handle, `mapping` is an
    iterable of 2-tuples (entity_handle, sort_handle) or a
    dict (entity_handle, sort_handle). Entities with equal sort handles show
    up in reversed source entities order.

    Args:
        entities: iterable of :class:`DXFGraphic` objects
        mapping: iterable of 2-tuples (entity_handle, sort_handle) or a
            handle mapping as dict.

    """
    mapping = dict(mapping) if mapping else {}
    heap = _build(entities, mapping, -1)
    return _sorted(heap)


def _sorted(heap) -> Iterable["DXFGraphic"]:
    """Yields heap content in order."""
    while heap:
        yield heapq.heappop(heap)[-1]


def _build(
    entities: Iterable["DXFGraphic"], mapping: Dict, order: int
) -> List[Tuple[int, int, "DXFGraphic"]]:
    """Returns a heap structure.

    Args:
        entities: DXF entities to order
        mapping: handle remapping
        order: +1 for ascending, -1 for descending

    """

    def sort_handle(entity: "DXFGraphic") -> int:
        handle = entity.dxf.handle
        sort_handle_ = mapping.get(handle, handle)
        if sort_handle_ == NULL_HANDLE:
            # This behavior is defined by AutoCAD but not documented in the
            # DXF reference.
            sort_handle_ = MAX_HANDLE
        return int(sort_handle_, 16)

    heap: List[Tuple[int, int, "DXFGraphic"]] = []
    for index, entity in enumerate(entities):
        # DXFGraphic is not sortable, using the index as second value avoids
        # a key function and preserves explicit the source order of
        # equal sort handles.
        heapq.heappush(
            heap,
            (
                sort_handle(entity) * order,
                index * order,
                entity,
            ),
        )
    return heap



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, List
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER, VERTEXNAMES
from ezdxf.math import Matrix44, Z_AXIS, NULLVEC, Vec3
from ezdxf.math.transformtools import OCSTransform
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity, elevation_to_z_axis
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace

__all__ = ["Solid", "Trace", "Face3d"]

acdb_trace = DefSubclass(
    "AcDbTrace",
    {
        # IMPORTANT: all 4 vertices have to be present in the DXF file,
        # otherwise AutoCAD shows a DXF structure error and does not load the
        # file! (SOLID, TRACE and 3DFACE)
        # 1. corner Solid WCS; Trace OCS
        "vtx0": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        # 2. corner Solid WCS; Trace OCS
        "vtx1": DXFAttr(11, xtype=XType.point3d, default=NULLVEC),
        # 3. corner Solid WCS; Trace OCS
        "vtx2": DXFAttr(12, xtype=XType.point3d, default=NULLVEC),
        # 4. corner Solid WCS; Trace OCS:
        # If only three corners are entered to define the SOLID, then the fourth
        # corner coordinate is the same as the third.
        "vtx3": DXFAttr(13, xtype=XType.point3d, default=NULLVEC),
        # IMPORTANT: for TRACE and SOLID the last two vertices are in reversed
        # order: a square has the vertex order 0-1-3-2
        # Elevation is a legacy feature from R11 and prior, do not use this
        # attribute, store the entity elevation in the z-axis of the vertices.
        # ezdxf does not export the elevation attribute!
        "elevation": DXFAttr(38, default=0, optional=True),
        # Thickness could be negative:
        "thickness": DXFAttr(39, default=0, optional=True),
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_trace_group_codes = group_code_mapping(acdb_trace)


class _Base(DXFGraphic):
    def __getitem__(self, num):
        return self.dxf.get(VERTEXNAMES[num])

    def __setitem__(self, num, value):
        return self.dxf.set(VERTEXNAMES[num], value)


@register_entity
class Solid(_Base):
    """DXF SHAPE entity"""

    DXFTYPE = "SOLID"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_trace)

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        """Loading interface. (internal API)"""
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_trace_group_codes, subclass=2, recover=True
            )
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, VERTEXNAMES)
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags. (internal API)"""
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_trace.name)
        if not self.dxf.hasattr("vtx3"):
            self.dxf.vtx3 = self.dxf.vtx2
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "vtx0",
                "vtx1",
                "vtx2",
                "vtx3",
                "thickness",
                "extrusion",
            ],
        )

    def transform(self, m: Matrix44) -> "Solid":
        """Transform the SOLID/TRACE entity by transformation matrix `m` inplace."""
        # SOLID and TRACE are OCS entities.
        dxf = self.dxf
        ocs = OCSTransform(self.dxf.extrusion, m)
        for name in VERTEXNAMES:
            if dxf.hasattr(name):
                dxf.set(name, ocs.transform_vertex(dxf.get(name)))
        if dxf.hasattr("thickness"):
            dxf.thickness = ocs.transform_thickness(dxf.thickness)
        dxf.extrusion = ocs.new_extrusion
        self.post_transform(m)
        return self

    def wcs_vertices(self, close: bool = False) -> List[Vec3]:
        """Returns WCS vertices in correct order,
        if argument `close` is ``True``, last vertex == first vertex.
        Does **not** return duplicated last vertex if represents a triangle.

        .. versionadded:: 0.15

        """
        ocs = self.ocs()
        return list(ocs.points_to_wcs(self.vertices(close)))

    def vertices(self, close: bool = False) -> List[Vec3]:
        """Returns OCS vertices in correct order,
        if argument `close` is ``True``, last vertex == first vertex.
        Does **not** return duplicated last vertex if represents a triangle.

        .. versionadded:: 0.15

        """
        dxf = self.dxf
        vertices: List[Vec3] = [dxf.vtx0, dxf.vtx1, dxf.vtx2]
        if dxf.vtx3 != dxf.vtx2:  # face is not a triangle
            vertices.append(dxf.vtx3)

        # adjust weird vertex order of SOLID and TRACE:
        # 0, 1, 2, 3 -> 0, 1, 3, 2
        if len(vertices) > 3:
            vertices[2], vertices[3] = vertices[3], vertices[2]

        if close and not vertices[0].isclose(vertices[-1]):
            vertices.append(vertices[0])
        return vertices


@register_entity
class Trace(Solid):
    """DXF TRACE entity"""

    DXFTYPE = "TRACE"


acdb_face = DefSubclass(
    "AcDbFace",
    {
        # 1. corner WCS:
        "vtx0": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        # 2. corner WCS:
        "vtx1": DXFAttr(11, xtype=XType.point3d, default=NULLVEC),
        # 3. corner WCS:
        "vtx2": DXFAttr(12, xtype=XType.point3d, default=NULLVEC),
        # 4. corner WCS:
        # If only three corners are entered to define the SOLID, then the fourth
        # corner coordinate is the same as the third.
        "vtx3": DXFAttr(13, xtype=XType.point3d, default=NULLVEC),
        # invisible:
        # 1 = First edge is invisible
        # 2 = Second edge is invisible
        # 4 = Third edge is invisible
        # 8 = Fourth edge is invisible
        "invisible": DXFAttr(70, default=0, optional=True),
    },
)
acdb_face_group_codes = group_code_mapping(acdb_face)


@register_entity
class Face3d(_Base):
    """DXF 3DFACE entity"""
    # IMPORTANT: for 3DFACE the last two vertices are in regular order:
    # a square has the vertex order 0-1-2-3

    DXFTYPE = "3DFACE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_face)

    def is_invisible_edge(self, num: int) -> bool:
        """Returns True if edge `num` is an invisible edge."""
        if num < 0 or num > 4:
            raise ValueError(f'invalid edge: {num}')
        return bool(self.dxf.invisible & (1 << num))

    def set_edge_visibility(self, num: int, visible: bool = False) -> None:
        """Set visibility of edge `num`, status `True` for visible, status
        `False` for invisible.
        """
        if num < 0 or num >= 4:
            raise ValueError(f'invalid edge: {num}')
        if not visible:
            self.dxf.invisible = self.dxf.invisible | (1 << num)
        else:
            self.dxf.invisible = self.dxf.invisible & ~(1 << num)

    def get_edges_visibility(self) -> List[bool]:
        # if the face is a triangle, a fourth visibility flag
        # may be present but is ignored
        return [not self.is_invisible_edge(i) for i in range(4)]

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_face_group_codes, subclass=2, recover=True
            )
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_face.name)
        if not self.dxf.hasattr("vtx3"):
            self.dxf.vtx3 = self.dxf.vtx2
        self.dxf.export_dxf_attribs(
            tagwriter, ["vtx0", "vtx1", "vtx2", "vtx3", "invisible"]
        )

    def transform(self, m: Matrix44) -> "Face3d":
        """Transform the 3DFACE  entity by transformation matrix `m` inplace."""
        dxf = self.dxf
        # 3DFACE is a real 3d entity
        dxf.vtx0, dxf.vtx1, dxf.vtx2, dxf.vtx3 = m.transform_vertices(
            (dxf.vtx0, dxf.vtx1, dxf.vtx2, dxf.vtx3)
        )
        self.post_transform(m)
        return self

    def wcs_vertices(self, close: bool = False) -> List[Vec3]:
        """Returns WCS vertices, if argument `close` is ``True``,
        last vertex == first vertex.

        returns 4 vertices when close=False and 5 vertices when close=True.
        Some edges may have 0 length.

        Compatibility interface to SOLID and TRACE. The 3DFACE entity returns
        already WCS vertices.
        """
        dxf = self.dxf
        vertices: List[Vec3] = [dxf.vtx0, dxf.vtx1, dxf.vtx2]
        vtx3 = dxf.get("vtx3")
        if isinstance(vtx3, Vec3) and vtx3 != dxf.vtx2:  # face is not a triangle
            vertices.append(vtx3)
        if close:
            vertices.append(vertices[0])
        return vertices



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import Iterable, TYPE_CHECKING, Tuple
from typing_extensions import Protocol, runtime_checkable
from ezdxf.query import EntityQuery

if TYPE_CHECKING:
    from ezdxf.entities import DXFGraphic, DXFEntity


# Protocol implemented for:
# - DXFTagStorage
# - ACAD_PROXY_ENTITY
# - INSERT
# - DIMENSION
# - LEADER
# - MLINE
# - ProxyGraphic
# - DXFGraphicProxy (drawing add-on)
#
# TODO: MLEADER, virtual_entities() not implemented yet


@runtime_checkable
class SupportsVirtualEntities(Protocol):
    """The virtual entities protocol is used to disassemble complex entities
    into DXF primitives like LINE, ARC, ... as REQUIREMENT to render these
    entities. Which means the entity does not have :func:`ezdxf.path.make_path`
    support, except the text entities TEXT, ATTRIB and MTEXT.

    Optional DECONSTRUCTION of entities into DXF primitives like LWPOLYLINE
    into LINE and ARC entities is NOT the intended usage of this protocol!

    This protocol is for consistent internal usage and does not replace
    the :meth:`virtual_entities` methods!

    """

    def __virtual_entities__(self) -> Iterable["DXFGraphic"]:
        ...


def virtual_entities(entity: SupportsVirtualEntities) -> Iterable["DXFGraphic"]:
    if isinstance(entity, SupportsVirtualEntities):
        return entity.__virtual_entities__()
    else:
        raise TypeError(
            f"{type(entity)!r} does not support the __virtual_entities__ protocol"
        )


def query_virtual_entities(entity: SupportsVirtualEntities) -> EntityQuery:
    return EntityQuery(virtual_entities(entity))


@runtime_checkable
class ReferencedBlocks(Protocol):
    def __referenced_blocks__(self) -> Iterable[str]:
        """Returns the handles to the BLOCK_RECORD entities for all BLOCK
        definitions used by an entity.
        """
        ...


_EMPTY_TUPLE: Tuple = tuple()


def referenced_blocks(entity: "DXFEntity") -> Iterable[str]:
    """Returns the handles to the BLOCK_RECORD entities for all BLOCK
    definitions used by an entity.
    """
    if isinstance(entity, ReferencedBlocks):
        return entity.__referenced_blocks__()
    else:
        return _EMPTY_TUPLE
